import unittest

class TestCatalogIndex(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.indexes.common import CatalogIndex
        return CatalogIndex

    def test_ctor(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        def callback(object, default):
            """ """
        index = Test(callback)
        index._docids = set()
        self.assertEqual(index.discriminator, callback)

    def test_ctor_callback(self):
        klass = self._getTargetClass()
        def _discriminator(obj, default):
            """ """
        class Test(klass, DummyIndex):
            pass
        index = Test(_discriminator)
        index._docids = set()
        self.assertTrue(index.discriminator is _discriminator)

    def test_ctor_string(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        self.assertEqual(index.discriminator, 'abc')

    def test_ctor_bad_discrim(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        self.assertRaises(ValueError, Test, None)

    def test_not_implemented_applies_methods(self):
        index = self._getTargetClass()('foo')
        for name in [
            'applyContains',
            'applyDoesNotContain',
            'applyEq',
            'applyNotEq',
            'applyGt',
            'applyLt',
            'applyGe',
            'applyLe',
            'applyAny',
            'applyNotAny',
            'applyAll',
            'applyNotAll',
            'applyInRange',
            'applyNotInRange']:
            self.assertRaises(NotImplementedError, getattr(index, name))

    def test_index_doc_callback_returns_nondefault(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        def callback(ob, default):
            return ob
        index = Test(callback)
        index._docids = set()
        self.assertEqual(index.index_doc(1, 'abc'), 'abc')
        self.assertEqual(index.value, 'abc')
        self.assertEqual(set(index.docids()), set([1]))

    def test_index_doc_string_discrim(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            abc = 'abc'
        dummy = Dummy()
        self.assertEqual(index.index_doc(1, dummy), 'abc')
        self.assertEqual(index.value, 'abc')
        self.assertEqual(set(index.docids()), set([1]))

    def test_index_doc_missing_value_unindexes(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = 'abc'
        self.assertEqual(index.index_doc(1, dummy), 'abc')
        del dummy.abc
        del index.value
        self.assertEqual(index.index_doc(1, dummy), None)
        self.assertEqual(set(index.docids()), set([1]))
        self.assertEqual(index.value, None)
        self.assertEqual(index.unindexed, 1)

    def test_index_doc_missing_value_then_with_value(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        self.assertEqual(index.index_doc(20, dummy), None)
        self.assertTrue(20 in index.docids())
        dummy.abc = 'foo'
        self.assertEqual(index.index_doc(20, dummy), 'foo')
        self.assertTrue(20 in index.docids())

    def test_index_doc_missing_value_then_unindex(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        self.assertEqual(index.index_doc(20, dummy), None)
        self.assertTrue(20 in index.docids())
        index.unindex_doc(20)
        self.assertFalse(20 in index.docids())

    def test_docids_with_indexed_and_not_indexed(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        self.assertEqual(index.index_doc(20, dummy), None)
        dummy.abc = 'foo'
        self.assertEqual(index.index_doc(21, dummy), 'foo')
        self.assertEqual(set(index.docids()), set([20, 21]))

    def test_index_doc_persistent_value_raises(self):
        from persistent import Persistent
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = Persistent()
        self.assertRaises(ValueError, index.index_doc, 1, dummy)

    def test_index_doc_broken_object_raises(self):
        from ZODB.broken import Broken
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = Broken()
        self.assertRaises(ValueError, index.index_doc, 1, dummy)

    def test_reindex_doc(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = set()
        class Dummy:
            abc = 'abc'
        dummy = Dummy()
        index.index_doc(1, dummy)
        index.reindex_doc(1, dummy)
        self.assertEqual(index.unindexed, 1)
        self.assertEqual(index.value, 'abc')
        self.assertEqual(set(index.docids()), set([1]))

    def test_migrate_to_0_8_0(self):
        klass = self._getTargetClass()
        class Test(klass, DummyIndex):
            pass
        index = Test('abc')
        index._docids = index.family.IF.Set([1, 2])
        self.assertEqual(set(index.docids()), set([1, 2]))
        all_docids = set([1, 2, 3, 4])
        index._migrate_to_0_8_0(all_docids)
        self.assertEqual(set(index.docids()), set([1, 2, 3, 4]))


from zerodbext.catalog.interfaces import ICatalogIndex
from zope.interface import implementer


@implementer(ICatalogIndex)
class DummyIndex(object):

    value = None

    def index_doc(self, docid, value):
        self._docids.add(docid)
        self.value = value
        return value

    def unindex_doc(self, docid):
        if docid in self._docids:
            self._docids.remove(docid)
            self.unindexed = docid

    def _indexed(self):
        return self._docids




import unittest

_marker = object()

class CatalogPathIndex2Tests(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.indexes.path2 import CatalogPathIndex2
        return CatalogPathIndex2

    def _makeOne(self, values=None,
                 discriminator='path', attr_discriminator=_marker):
        if values is None:
            values = {}
        def _attr_discriminator(obj, default):
            return obj.path
        if attr_discriminator is _marker:
            attr_discriminator = _attr_discriminator
        index = self._getTargetClass()(discriminator,
                                       attr_discriminator=attr_discriminator)
        for doc_id, path in values.items():
            index.index_doc(doc_id, path)
        return index

    def test_class_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyClass
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyClass(ICatalogIndex, self._getTargetClass())

    def test_instance_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyObject
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyObject(ICatalogIndex, self._makeOne())

    def test_ctor_callback_discriminator(self):
        def _discriminator(obj, default):
            """ """
        index = self._makeOne(discriminator=_discriminator)
        self.assertTrue(index.discriminator is _discriminator)

    def test_ctor_string_discriminator(self):
        index = self._makeOne(discriminator='abc')
        self.assertEqual(index.discriminator, 'abc')

    def test_ctor_bad_discriminator(self):
        self.assertRaises(ValueError, self._makeOne, discriminator=object())

    def test_ctor_callback_attr_discriminator(self):
        def _attr_discriminator(obj, default):
            """ """
        index = self._makeOne(attr_discriminator=_attr_discriminator)
        self.assertTrue(index.attr_discriminator is _attr_discriminator)

    def test_ctor_string_attr_discriminator(self):
        index = self._makeOne(attr_discriminator='abc')
        self.assertEqual(index.attr_discriminator, 'abc')

    def test_ctor_no_attr_discriminator(self):
        index = self._makeOne(attr_discriminator=None)
        self.assertEqual(index.attr_discriminator, None)

    def test_ctor_bad_attr_discriminator(self):
        self.assertRaises(ValueError, self._makeOne,
                          attr_discriminator=object())

    def test_throws_on_no_query(self):
        index = self._makeOne({})
        self.assertRaises(KeyError, index.apply, {})

    def test_empty_index(self):
        index = self._makeOne({})
        self.assertEqual(len(index), 0)

    def test_nonempty_index(self):
        index = self._makeOne(VALUES)
        self.assertEqual(len(index), 22)
        self.assertTrue(index)

    def test_index_object_bad_path(self):
        index = self._makeOne()
        class Dummy:
            path = ()
        self.assertRaises(ValueError, index.index_doc, 1, Dummy())

    def test_index_object_simple(self):
        index = self._makeOne()
        class Dummy:
            path = '/abc'
        index.index_doc(1, Dummy())
        self.assertEqual(index.path_to_docid[('', 'abc')], 1)
        self.assertEqual(index.docid_to_path[1], ('', 'abc'))

    def test_index_object_callback_discriminator(self):
        def _discriminator(obj, default):
            return '/abc'
        index = self._makeOne(discriminator=_discriminator)
        class Dummy:
            path = '/foo'
        index.index_doc(1, Dummy())
        self.assertEqual(index.path_to_docid[('', 'abc')], 1)
        self.assertEqual(index.docid_to_path[1], ('', 'abc'))

    def test_index_object_missing_value_unindexes(self):
        index = self._makeOne()
        class Dummy:
            pass
        dummy = Dummy()
        dummy.path = '/abc'
        index.index_doc(1, Dummy())
        del dummy.path
        index.index_doc(1, Dummy())
        self.assertFalse(('', 'abc') in index.path_to_docid)
        self.assertFalse(1 in index.docid_to_path)

    def test_unindex_nonesuch(self):
        index = self._makeOne({})
        index.unindex_doc(1234) # nothrow

    def test_unindex_doc(self):
        index = self._makeOne(VALUES)
        docids = list(VALUES.keys())

        for doc_id in docids:
            index.unindex_doc(doc_id)

        self.assertEqual(len(index), 0)
        self.assertEqual(list(index.adjacency.keys()), [])
        self.assertEqual(list(index.disjoint.keys()), [])
        self.assertEqual(list(index.path_to_docid.keys()), [])
        self.assertEqual(list(index.docid_to_path.keys()), [])

        index = self._makeOne(VALUES)
        # randomize the order
        import random
        random.shuffle(docids)
        for doc_id in docids:
            index.unindex_doc(doc_id)

        self.assertEqual(len(index), 0)
        self.assertEqual(list(index.adjacency.keys()), [])
        self.assertEqual(list(index.disjoint.keys()), [])
        self.assertEqual(list(index.path_to_docid.keys()), [])
        self.assertEqual(list(index.docid_to_path.keys()), [])

    def test_unindex_then_index_doesnt_dupe(self):
        index = self._makeOne()
        o = Dummy('/foo/bar')
        index.index_doc(1, o)
        self.assertEqual(len(index), 1)
        index.index_doc(1, o)
        self.assertEqual(len(index), 1)

    # This test fails, but I can't see why it should.
    #def XXX_test_search_from_unindexed_root(self):
    #    index = self._makeOne()
    #    index.index_doc(1, '/abc/def')
    #    index.index_doc(2, '/abc/ghi')
    #    result = index.search('/abc', depth=1, include_path=True)
    #    self.assertEqual(sorted(result), [1, 2])

    def test_search_root_nodepth(self):
        index = self._makeOne(VALUES)
        result = index.search('/')
        self.assertEqual(sorted(result), list(range(1, 21)))

    def test_search_root_nodepth_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/', include_path=True)
        self.assertEqual(sorted(result), list(range(0, 21)))

    def test_search_root_depth_0(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=0)
        self.assertEqual(sorted(result), [])

    def test_search_depth_0_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=0, include_path=True)
        self.assertEqual(sorted(result), [0])

    def test_search_root_depth_1(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=1)
        self.assertEqual(sorted(result), [1,5])

    def test_search_depth_1_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=1, include_path=True)
        self.assertEqual(sorted(result), [0, 1, 5])

    def test_search_root_depth_2(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=2)
        self.assertEqual(sorted(result), [1, 2, 3, 4, 5, 6, 7, 8])

    def test_search_depth_2_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=2, include_path=True)
        self.assertEqual(sorted(result), [0, 1, 2, 3, 4, 5, 6, 7, 8])

    def test_search_root_depth_3(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=3)
        self.assertEqual(sorted(result), list(range(1, 21)))

    def test_search_root_depth_3_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/', depth=3, include_path=True)
        self.assertEqual(sorted(result), list(range(0, 21)))

    def test_search_aa_nodepth(self):
        index = self._makeOne(VALUES)
        result = index.search('/aa')
        self.assertEqual(sorted(result), [2, 3, 4, 9, 10, 11, 12, 13, 14])

    def test_search_aa_nodepth_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/aa', include_path=True)
        self.assertEqual(sorted(result), [1, 2, 3, 4, 9, 10, 11, 12, 13, 14])

    def test_search_aa_depth_0(self):
        index = self._makeOne(VALUES)
        result = index.search('/aa', depth=0)
        self.assertEqual(sorted(result), [])

    def test_search_aa_depth_0_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/aa', depth=0, include_path=True)
        self.assertEqual(sorted(result), [1])

    def test_search_aa_depth_1(self):
        index = self._makeOne(VALUES)
        result = index.search('/aa', depth=1)
        self.assertEqual(sorted(result), [2,3,4])

    def test_search_aa_depth_1_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/aa', depth=1, include_path=True)
        self.assertEqual(sorted(result), [1, 2, 3, 4])

    def test_search_bb_nodepth(self):
        index = self._makeOne(VALUES)
        result = index.search('/bb')
        self.assertEqual(sorted(result), [6, 7, 8, 15, 16, 17, 18, 19, 20])

    def test_search_bb_nodepth_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/bb', include_path=True)
        self.assertEqual(sorted(result), [5, 6, 7, 8, 15, 16, 17, 18, 19, 20])

    def test_search_bb_depth_0(self):
        index = self._makeOne(VALUES)
        result = index.search('/bb', depth=0)
        self.assertEqual(sorted(result), [])

    def test_search_bb_depth_0_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/bb', depth=0, include_path=True)
        self.assertEqual(sorted(result), [5])

    def test_search_bb_depth_1(self):
        index = self._makeOne(VALUES)
        result = index.search('/bb', depth=1)
        self.assertEqual(sorted(result), [6, 7, 8])

    def test_search_bb_depth_1_include_path(self):
        index = self._makeOne(VALUES)
        result = index.search('/bb', depth=1, include_path=True)
        self.assertEqual(sorted(result), [5, 6, 7, 8])

    def test_search_with_tuple(self):
        index = self._makeOne(VALUES)
        result = index.search(('', 'bb'))
        self.assertEqual(sorted(result), [6, 7, 8, 15, 16, 17, 18, 19, 20])

    def test_disjoint_resolved(self):
        index = self._makeOne(VALUES)
        result = index.search('/disjoint')
        self.assertEqual(sorted(result), [])
        index.index_doc(22, Dummy('/disjoint'))
        result = index.search('/disjoint')
        self.assertEqual(sorted(result), [21])
        result = index.search('/')
        self.assertTrue(22 in result)

    def test_disjoint_resolved_random_order(self):
        index = self._makeOne({})

        index.index_doc(0, Dummy('/disjoint/path/element/1'))
        index.index_doc(1, Dummy('/disjoint/path/element/2'))

        result = index.search('/')
        self.assertEqual(sorted(result), [])

        index.index_doc(2, Dummy('/disjoint'))

        result = index.search('/')
        self.assertEqual(sorted(result), [])

        index.index_doc(3, Dummy('/disjoint/path'))

        result = index.search('/')
        self.assertEqual(sorted(result), [])

        index.index_doc(4, Dummy('/'))

        result = index.search('/')
        self.assertEqual(sorted(result), [2, 3])

        index.index_doc(5, Dummy('/disjoint/path/element'))

        result = index.search('/')
        self.assertEqual(sorted(result), [0, 1, 2, 3, 5])

    def test_apply_intersect_wo_docids(self):
        index = self._makeOne(VALUES)
        result = index.apply_intersect('/aa', None)
        self.assertEqual(sorted(result), [2, 3, 4, 9, 10, 11, 12, 13, 14])

    def test_apply_intersect_w_docids(self):
        index = self._makeOne(VALUES)
        to_intersect = index.family.IF.Set([2, 3, 4, 7, 8])
        result = index.apply_intersect('/aa', to_intersect)
        self.assertEqual(sorted(result), [2, 3, 4])

    def test_apply_path_string(self):
        index = self._makeOne(VALUES)
        result = index.apply('/aa')
        self.assertEqual(sorted(result), [2, 3, 4, 9, 10, 11, 12, 13, 14])

    def test_apply_path_string_relative_raises(self):
        index = self._makeOne(VALUES)
        result = self.assertRaises(ValueError, index.apply, 'aa')

    def test_apply_path_tuple_absolute(self):
        index = self._makeOne(VALUES)
        result = index.apply(('', 'aa'))
        self.assertEqual(sorted(result), [2, 3, 4, 9, 10, 11, 12, 13, 14])

    def test_apply_path_tuple_relative_raises(self):
        index = self._makeOne(VALUES)
        result = self.assertRaises(ValueError, index.apply, ('aa',))

    def test_apply_path_list_absolute(self):
        index = self._makeOne(VALUES)
        result = index.apply(['', 'aa'])
        self.assertEqual(sorted(result), [2, 3, 4, 9, 10, 11, 12, 13, 14])

    def test_apply_path_list_relative_raises(self):
        index = self._makeOne(VALUES)
        result = self.assertRaises(ValueError, index.apply, ['aa'])

    def test_apply_path_dict_with_string_query(self):
        index = self._makeOne(VALUES)
        result = index.apply({'query':'/aa', 'depth':1})
        self.assertEqual(sorted(result), [2, 3, 4])

    def test_apply_path_dict_with_tuple_query(self):
        index = self._makeOne(VALUES)
        result = index.apply({'query':('', 'aa'), 'depth':1})
        self.assertEqual(sorted(result), [2, 3, 4])

    def test_apply_path_dict_with_list_query(self):
        index = self._makeOne(VALUES)
        result = index.apply({'query':['', 'aa'], 'depth':1})
        self.assertEqual(sorted(result), [2, 3, 4])

    def test_apply_path_dict_with_include_path_true(self):
        index = self._makeOne(VALUES)
        result = index.apply({'query':('', 'aa'), 'depth':1,
                              'include_path':True})
        self.assertEqual(sorted(result), [1, 2, 3, 4])

    def test_apply_path_dict_with_include_path_false(self):
        index = self._makeOne(VALUES)
        result = index.apply({'query':('', 'aa'), 'depth':1,
                              'include_path':False})
        self.assertEqual(sorted(result), [2, 3, 4])

    def test_reindex_doc_nochange(self):
        index = self._makeOne(VALUES)
        result = index.reindex_doc(1, VALUES[1])
        self.assertEqual(result, False)

    def test_reindex_doc_attradd(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        self.assertEqual(index.docid_to_attr.get(1), None)
        result = index.reindex_doc(1, Dummy('/aa', 'acl5'))
        self.assertEqual(result, True)
        self.assertEqual(index.docid_to_attr[1], 'acl5')

    def test_reindex_doc_attrchange(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        self.assertEqual(index.docid_to_attr[0], 'acl1')
        result = index.reindex_doc(0, Dummy('/', 'acl11'))
        self.assertEqual(result, True)
        self.assertEqual(index.docid_to_attr[0], 'acl11')

    def test_reindex_doc_attrdel(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        self.assertEqual(index.docid_to_attr[0], 'acl1')
        result = index.reindex_doc(0, Dummy('/'))
        self.assertEqual(result, True)
        self.assertEqual(index.docid_to_attr.get(0), None)

    def test_reindex_doc_pathchange(self):
        index = self._makeOne(VALUES)
        result = index.reindex_doc(1, Dummy('/abcdef'))
        self.assertEqual(result, True)
        self.assertEqual(index.docid_to_path[1], ('', 'abcdef'))

    def test_index_doc_unindex_doc_with_attr_discriminator(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_keys = sorted(index.docid_to_attr.keys())
        self.assertEqual(attr_keys, [0, 2, 8, 12])

        docids = list(VALUES.keys())

        for doc_id in docids:
            index.unindex_doc(doc_id)

        self.assertEqual(len(index), 0)
        self.assertEqual(list(index.adjacency.keys()), [])
        self.assertEqual(list(index.disjoint.keys()), [])
        self.assertEqual(list(index.path_to_docid.keys()), [])
        self.assertEqual(list(index.docid_to_path.keys()), [])
        self.assertEqual(list(index.docid_to_attr.keys()), [])

        index = self._makeOne(VALUES, attr_discriminator='attr')
        # randomize the order
        import random
        random.shuffle(docids)
        for doc_id in docids:
            index.unindex_doc(doc_id)

        self.assertEqual(len(index), 0)
        self.assertEqual(list(index.adjacency.keys()), [])
        self.assertEqual(list(index.disjoint.keys()), [])
        self.assertEqual(list(index.path_to_docid.keys()), [])
        self.assertEqual(list(index.docid_to_path.keys()), [])
        self.assertEqual(list(index.docid_to_attr.keys()), [])

    def test_search_root_nodepth_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', attr_checker=attr_checker)
        self.assertEqual(sorted(result), list(range(1, 21)))
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 4)
        attrs, theset = results[3]
        self.assertEqual(attrs, (12, ['acl1', 'acl4']))
        self.assertEqual(sorted(theset), [12])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [1,3,4,5,6,7,11,13,14,15,16,17,18])
        attrs, theset = results[2]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(sorted(theset), [8, 19, 20])
        attrs, theset = results[1]
        self.assertEqual(attrs, (2, ['acl1', 'acl3']))
        self.assertEqual(sorted(theset), [2,9, 10])

    def test_search_root_nodepth_with_attrchecker_and_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', include_path=True, attr_checker=attr_checker)
        self.assertEqual(sorted(result), list(range(0, 21)))
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 4)
        attrs, theset = results[3]
        self.assertEqual(attrs, (12, ['acl1', 'acl4']))
        self.assertEqual(sorted(theset), [12])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0,1,3,4,5,6,7,11,13,14,15,16,17,18])
        attrs, theset = results[2]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(sorted(theset), [8, 19, 20])
        attrs, theset = results[1]
        self.assertEqual(attrs, (2, ['acl1', 'acl3']))
        self.assertEqual(sorted(theset), [2,9, 10])

    def test_search_root_depth0_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=0, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 1)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])

    def test_search_root_depth0_with_attrchecker_and_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=0, include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [0])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 1)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0])

    def test_search_root_depth1_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=1, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [1, 5])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 1)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [1,5])

    def test_search_root_depth1_with_attrchecker_and_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=1, include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [0, 1, 5])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 1)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0, 1, 5])

    def test_search_root_depth2_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=2, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [1, 2, 3, 4, 5, 6, 7, 8])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 3)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [1,3,4,5,6,7])
        attrs, theset = results[2]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8])
        attrs, theset = results[1]
        self.assertEqual(attrs, (2, ['acl1', 'acl3']))
        self.assertEqual(list(theset), [2])

    def test_search_root_depth2_with_attrchecker_and_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=2, include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [0, 1, 2, 3, 4, 5, 6, 7, 8])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 3)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0,1,3,4,5,6,7])
        attrs, theset = results[2]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8])
        attrs, theset = results[1]
        self.assertEqual(attrs, (2, ['acl1', 'acl3']))
        self.assertEqual(list(theset), [2])


    def test_search_root_depth3_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=3, attr_checker=attr_checker)
        self.assertEqual(sorted(result), list(range(1, 21)))
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 4)
        attrs, theset = results[3]
        self.assertEqual(attrs, (12, ['acl1', 'acl4']))
        self.assertEqual(list(theset), [12])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [1,3,4,5,6,7,11,13,14,15,16,17,18])
        attrs, theset = results[2]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8, 19, 20])
        attrs, theset = results[1]
        self.assertEqual(attrs, (2, ['acl1', 'acl3']))
        self.assertEqual(list(theset), [2, 9, 10])

    def test_search_root_depth3_with_attrchecker_and_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=3, include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), list(range(0, 21)))
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 4)
        attrs, theset = results[3]
        self.assertEqual(attrs, (12, ['acl1', 'acl4']))
        self.assertEqual(list(theset), [12])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0,1,3,4,5,6,7,11,13,14,15,16,17,18])
        attrs, theset = results[2]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8, 19, 20])
        attrs, theset = results[1]
        self.assertEqual(attrs, (2, ['acl1', 'acl3']))
        self.assertEqual(list(theset), [2, 9, 10])

    def test_search_bb_nodepth_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb', attr_checker=attr_checker)
        self.assertEqual(sorted(result), [6, 7, 8, 15, 16, 17, 18, 19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [6, 7, 15, 16, 17, 18])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8, 19, 20])

    def test_search_bb_nodepth_with_attrchecker_and_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb', include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [5, 6, 7, 8, 15, 16, 17, 18, 19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [5, 6, 7, 15, 16, 17, 18])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8, 19, 20])

    def test_search_bb_depth0_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb', depth=0, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 1)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])

    def test_search_bb_depth1_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb', depth=1, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [6,7,8])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [6,7])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8])

    def test_search_bb_depth2_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb', depth=2, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [6, 7, 8, 15, 16, 17, 18, 19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [6, 7, 15, 16, 17, 18])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8, 19, 20])

    def test_search_bb_depth3_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb', depth=3, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [6, 7, 8, 15, 16, 17, 18, 19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [6, 7, 15, 16, 17, 18])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8, 19, 20])

    def test_search_bb_cc_nodepth_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc', attr_checker=attr_checker)
        self.assertEqual(sorted(result), [19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [19, 20])

    def test_search_bb_cc_nodepth_with_attrchecker_and_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc', include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [8, 19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [8, 19, 20])

    def test_search_bb_cc_depth0_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc', depth=0, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [])

    def test_search_bb_cc_depth1_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc', depth=1, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [19, 20])

    def test_search_bb_cc_depth2_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc', depth=2, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [19, 20])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [19, 20])

    def test_search_bb_cc_11_nodepth_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc/11.html', attr_checker=attr_checker)
        self.assertEqual(sorted(result), [])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [])

    def test_search_bb_cc_11_nodepth_with_attrchecker_include_path(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc/11.html', include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [19])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [19])

    def test_search_bb_cc_11_depth0_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc/11.html', depth=0,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [])

    def test_search_bb_cc_11_depth1_with_attrchecker(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc/11.html', depth=1,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [])

    def test_search_bb_cc_11_depth1_with_attrchecker_and_includepath(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/bb/cc/11.html', depth=1, include_path=True,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [19])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 2)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])
        attrs, theset = results[1]
        self.assertEqual(attrs, (8, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [19])

    def test_search_root_nodepth_with_attrchecker_values2(self):
        index = self._makeOne(VALUES2, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', attr_checker=attr_checker)
        self.assertEqual(sorted(result), list(range(1, 6)))
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 6)
        attrs, theset = results[1]
        self.assertEqual(attrs, (1, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [1])
        attrs, theset = results[3]
        self.assertEqual(attrs, (3, ['acl1', 'acl2', 'acl4']))
        self.assertEqual(list(theset), [3])
        attrs, theset = results[4]
        self.assertEqual(attrs, (4, ['acl1', 'acl2', 'acl5']))
        self.assertEqual(list(theset), [4])
        attrs, theset = results[2]
        self.assertEqual(attrs, (2, ['acl1', 'acl2', 'acl3']))
        self.assertEqual(list(theset), [2])
        attrs, theset = results[5]
        self.assertEqual(attrs, (5, ['acl1', 'acl6']))
        self.assertEqual(list(theset), [5])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])

    def test_search_root_nodepth_with_attrchecker_incpath_values2(self):
        index = self._makeOne(VALUES2, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', include_path=True, attr_checker=attr_checker)
        self.assertEqual(sorted(result), list(range(0, 6)))
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 6)
        attrs, theset = results[1]
        self.assertEqual(attrs, (1, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [1])
        attrs, theset = results[3]
        self.assertEqual(attrs, (3, ['acl1', 'acl2', 'acl4']))
        self.assertEqual(list(theset), [3])
        attrs, theset = results[4]
        self.assertEqual(attrs, (4, ['acl1', 'acl2', 'acl5']))
        self.assertEqual(list(theset), [4])
        attrs, theset = results[2]
        self.assertEqual(attrs, (2, ['acl1', 'acl2', 'acl3']))
        self.assertEqual(list(theset), [2])
        attrs, theset = results[5]
        self.assertEqual(attrs, (5, ['acl1', 'acl6']))
        self.assertEqual(list(theset), [5])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0])

    def test_search_root_depth0_with_attrchecker_values2(self):
        index = self._makeOne(VALUES2, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=0, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 1)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])

    def test_search_root_depth0_with_attrchecker_incpath_values2(self):
        index = self._makeOne(VALUES2, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', include_path=True, depth=0,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [0])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 1)
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0])

    def test_search_root_depth1_with_attrchecker_values2(self):
        index = self._makeOne(VALUES2, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', depth=1, attr_checker=attr_checker)
        self.assertEqual(sorted(result), [1,5])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 3)
        attrs, theset = results[2]
        self.assertEqual(attrs, (5, ['acl1', 'acl6']))
        self.assertEqual(list(theset), [5])
        attrs, theset = results[1]
        self.assertEqual(attrs, (1, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [1])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [])

    def test_search_root_depth1_with_attrchecker_incpath_values2(self):
        index = self._makeOne(VALUES2, attr_discriminator='attr')
        attr_checker = DummyAttrChecker()
        result = index.search('/', include_path=True, depth=1,
                              attr_checker=attr_checker)
        self.assertEqual(sorted(result), [0, 1, 5])
        results = sorted(attr_checker.results)
        self.assertEqual(len(results), 3)
        attrs, theset = results[2]
        self.assertEqual(attrs, (5, ['acl1', 'acl6']))
        self.assertEqual(list(theset), [5])
        attrs, theset = results[1]
        self.assertEqual(attrs, (1, ['acl1', 'acl2']))
        self.assertEqual(list(theset), [1])
        attrs, theset = results[0]
        self.assertEqual(attrs, (0, ['acl1']))
        self.assertEqual(list(theset), [0])

    def test__indexed(self):
        index = self._makeOne(VALUES, attr_discriminator='attr')
        self.assertEqual(
            set(index._indexed()),
            set(range(22)))

class DummyAttrChecker(object):
    def __init__(self):
        pass

    def __call__(self, results):
        import BTrees
        self.results = results
        return BTrees.family32.IF.multiunion([x[1] for x in results])

class Dummy:

    def __init__( self, path, attr=None):
        self.path = path
        if attr is not None:
            self.attr = attr

VALUES = {
    0 : Dummy('/', 'acl1'),
    1 : Dummy('/aa'),
    2 : Dummy('/aa/aa', 'acl3'),
    3 : Dummy('/aa/bb'),
    4 : Dummy('/aa/cc'),
    5 : Dummy('/bb'),
    6 : Dummy('/bb/aa'),
    7 : Dummy('/bb/bb'),
    8 : Dummy('/bb/cc', 'acl2'),
    9 : Dummy("/aa/aa/1.html"),
    10 : Dummy("/aa/aa/2.html"),
    11 : Dummy("/aa/bb/3.html"),
    12 : Dummy("/aa/bb/4.html", 'acl4'),
    13 : Dummy("/aa/cc/5.html"),
    14 : Dummy("/aa/cc/6.html"),
    15 : Dummy("/bb/aa/7.html"),
    16 : Dummy("/bb/aa/8.html"),
    17 : Dummy("/bb/bb/9.html"),
    18 : Dummy("/bb/bb/10.html"),
    19 : Dummy("/bb/cc/11.html"),
    20 : Dummy("/bb/cc/12.html"),
    21 : Dummy('/disjoint/entry')
    }

VALUES2 = {
    0 : Dummy('/', 'acl1'),
    1 : Dummy('/aa', 'acl2'),
    2 : Dummy('/aa/aa', 'acl3'),
    3 : Dummy('/aa/bb', 'acl4'),
    4 : Dummy('/aa/cc', 'acl5'),
    5 : Dummy('/bb', 'acl6'),
    }



"""
Thinking about apply_intersection()

oob = 250  (OOBTree DEFAULT_MAX_BTREE_SIZE)
iob = 500  (IOBTree DEFAULT_MAX_BTREE_SIZE)
Given two indices: i1, i2
Number of docs in each index: nd1, nd2
Number of unique keys in each index: nk1, nk2
Average size of result set: nd1/nk1, nd2/nk2

Lookup on i1 has been performed, yielding result set r1 with actual size n1.
r2 is not yet computed, average n2 is nd2/nk2.

Two algorithms:

1) Perform lookup on i2, yielding r2, then take intersection of r1 and r2.

Fudge factor is introduced to account for fact that forward lookup is done with
string keys whereas reverse lookup and intersection are performed with integer
keys which have less expensive comparisons.

Lookup is O(fudge * log(nk2, oob))
Intersection is O(max(n1, n2))
Total cost: O(log(nk2, oob) + max(n1, n2))
or taking the average for n2, since not yet known,
Total cost: O(log(nk2, oob) + max(n1, nd2/nk2))

2) Iterate over items in r1, use rev_index of i2 to check for each value in i2.

Rev lookup in i2 is O(log(nd2, iob))
Performed n1 times
Total cost: O(n1 * log(nd2, iob))

2) beats 1) when:

  n1 * log(nd2, iob) < fudge * log(nk2, oob) + max(n1, nd2/nk2)

So given a relationship between number of keys and number of docs in i2 and
size of r1, we can decide which algorithm is better.

We can also look at order of evaluation of r1 vs r2. Given there is some
advantage to algorithm 2 if n1 is small, we can compute nd/nk for each index
and choose to evaluate the index with the smaller value first, since on
average it will have a smaller result set.

That's the theory anyway.  Let's see who wins.
"""
import glob
import math
import os
import random
import sys
import time

from zerodbext.catalog.catalog import ConnectionManager
from zerodbext.catalog.catalog import FileStorageCatalogFactory
from zerodbext.catalog.indexes.field import CatalogFieldIndex
from zerodbext.catalog.query import Eq
from zerodbext.catalog.query import BoolOp

_marker = object()
random.seed()

class Intersection1(BoolOp):
    """
    Total cost: O(log(nk2, oob) + max(n1, nd2/nk2))
    """
    def apply(self, catalog):
        left = self.left.apply(catalog)
        if len(left) == 0:
            results = self.family.IF.Set()
        else:
            right = self.right.apply(catalog)
            if len(right) == 0:
                results = self.family.IF.Set()
            else:
                _, results = self.family.IF.weightedIntersection(left, right)
        return results

class Intersection2(BoolOp):
    """
    Implements algorithm2 above.  In real life we wouldn't do this in the
    Intersection operator--we'd do it in the apply_intersection() of the index
    and wire the Intersection operator to use that.

    Total cost: O(n1 * log(nd2, iob))
    """
    def apply(self, catalog):
        left = self.left.apply(catalog)
        rev_index = catalog[self.right.index_name]._rev_index
        value = self.right.value
        results = self.family.IF.Set()
        for docid in left:
            if rev_index.get(docid, _marker) == value:
                results.add(docid)
        return results


def predictions(nd, nk1, nk2):
    FUDGE = 17.0
##    oob = 250 # OOBTree DEFAULT_MAX_BTREE_SIZE
##    iob = 500 # IOBTree DEFAULT_MAX_BTREE_SIZE
    oob = 125 # OOBTree wag avg bucket size
    iob = 250 # IOBTree wag avg bucket size
    L_FWD_LOOKUP_COST = FUDGE * math.log(nk1, oob)
    R_FWD_LOOKUP_COST = FUDGE * math.log(nk2, oob)
    L_REV_LOOKUP_COST = math.log(nd, iob)
    AVG_L_RESULT_SIZE = float(nd)/nk1
    AVG_R_RESULT_SIZE = float(nd)/nk2
    MAX_INTERSECT_COST = max(AVG_L_RESULT_SIZE, AVG_R_RESULT_SIZE)
    AVG_INTERSECT_COST = MAX_INTERSECT_COST / 2.0 #max(nk1/2, nk2/2) / 2

    # Total cost: O(log(nk2, oob) + max(n1, nd2/nk2))
    cost1 = L_FWD_LOOKUP_COST + R_FWD_LOOKUP_COST + AVG_INTERSECT_COST
    # Total cost: O(n1 * log(nd2, iob))
    cost2 = L_FWD_LOOKUP_COST + (AVG_L_RESULT_SIZE * L_REV_LOOKUP_COST)

    return cost1, cost2

##def predictions(nd, nk1, nk2):
##    s1 = nd / nk1
##    s2 = nd / nk2
##    if s1 <= s2 / 2:
##        return 2.0, 1.0
##    return 1.0, 2.0

def do_benchmark(fname, nd, nk1, nk2, out=sys.stdout):
    cumulative1 = 0.0
    cumulative2 = 0.0

    print >>out, "Index 1:"
    print >>out, "\t# docs: %d" % nd
    print >>out, "\t# distinct keys: %d" % nk1
    print >>out, "Index 2:"
    print >>out, "\t# docs: %d" % nd
    print >>out, "\t# distinct keys: %d" % nk2
    print >>out, ""

    cost1, cost2 = predictions(nd, nk1, nk2)

    print >>out, 'Cost1: %0.2f' % cost1
    print >>out, 'Cost2: %0.2f' % cost2
    print >>out
    print >>out, "Prediction:"
    if cost1 > cost2:
        print >>out, "Algorithm 2 %0.2f times faster than Algorithm 1" % (
            cost1/cost2)
    else:
        print >>out, "Algorithm 1 %0.2f times faster than Algorithm 2" % (
            cost2/cost1)

    print >>out, ""
    print >>out, "Setting up indexes..."
    for fn in glob.glob(fname + "*"):
        os.remove(fn)

    manager = ConnectionManager()
    factory = FileStorageCatalogFactory(fname, 'intersection')
    catalog = factory(manager)

    catalog['one'] = CatalogFieldIndex('one')
    catalog['two'] = CatalogFieldIndex('two')

    class Document(object):
        def __init__(self, docid):
            self.one = str(docid % nk1)
            self.two = str(docid % nk2)

    for docid in xrange(nd):
        catalog.index_doc(docid, Document(docid))
    manager.commit()
    manager.close()

    N_QUERIES = 1000
    print >>out, "Running %d queries for each algorithm..." % N_QUERIES
    catalog = factory(manager)
    for _ in xrange(1000):
        key1 = random.randrange(nk1)
        key2 = random.randrange(nk2)
        query1 = Intersection1(Eq('one', str(key1)), Eq('two', str(key2)))
        query2 = Intersection2(Eq('one', str(key1)), Eq('two', str(key2)))

        start = time.time()
        result1 = query1.apply(catalog)
        cumulative1 += time.time() - start

        start = time.time()
        result2 = query2.apply(catalog)
        cumulative2 += time.time() - start

        s1 = sorted(list(result1))
        s2 = sorted(list(result2))

        assert s1==s2, (s1, s2)

    manager.close()
    for fn in glob.glob(fname + "*"):
        os.remove(fn)

    print >>out, ""
    print >>out, "Result:"
    print >>out, "Time for algorithm1: %0.3f s" % cumulative1
    print >>out, "Time for algorithm2: %0.3f s" % cumulative2
    if cumulative1 > cumulative2:
        print >>out, "Algorithm 2 %0.2f times faster than Algorithm 1" % (
            cumulative1/cumulative2)
    else:
        print >>out, "Algorithm 1 %0.2f times faster than Algorithm 2" % (
            cumulative2/cumulative1)
    return cost1 / cost2, cumulative1 / cumulative2

class Null(object):
    def write(self, s):
        pass

def _range_order_of_magnitude(n):
    # Iterate over (at most) 3 orders of magnitude
    n_magnitude = int(math.ceil(math.log10(n)))
    lowest_magnitude = max(0, n_magnitude - 3)
    for magnitude in xrange(lowest_magnitude, n_magnitude):
        for i in xrange(1,10):
            value = i * 10**magnitude
            if value >= n:
                break
            yield value

def do_benchmarks(fname):
    null = Null()
    print "Cost of algorithm 1 / Cost of algorithm 2"
    print "N Docs | N Keys 1 | N Keys 2 | Predicted | Actual | Correct"
    for nd in  [100, 1000, 10000, 100000, 1000000]:
        for nk1 in _range_order_of_magnitude(nd / 2):
            for nk2 in _range_order_of_magnitude(nd):
                predicted, actual = do_benchmark(fname, nd, nk1, nk2, out=null)
                correct = ((predicted >= 1 and actual >= 1) or
                           (predicted < 1 and actual < 1))
                print "%6d | %8d | %8d | %9.2f | %6.2f | %s" % (
                    nd, nk1, nk2,  predicted, actual, correct)
                sys.stdout.flush()

# profile (unused right now)
def profile(cmd, globals, locals, sort_order, callers):
    import profile
    import pstats
    import tempfile
    fd, fn = tempfile.mkstemp()
    try:
        if hasattr(profile, 'runctx'):
            profile.runctx(cmd, globals, locals, fn)
        else:
            raise NotImplementedError('No profiling support under Python 2.3')
        stats = pstats.Stats(fn)
        stats.strip_dirs()
        # calls,time,cumulative and cumulative,calls,time are useful
        stats.sort_stats(*sort_order or ('cumulative', 'calls', 'time'))
        if callers:
            stats.print_callers(.3)
        else:
            stats.print_stats(.3)
    finally:
        os.remove(fn)

def rerun_predictions(fname):
    benchmarks = open(fname).xreadlines()
    benchmarks.next(); benchmarks.next()  # skip header lines

    print "Cost of algorithm 1 / Cost of algorithm 2"
    print "nd     | nd/nk1   | nd/nk2   | Predicted | Actual | Correct"

    gain = count = n_correct = 0
    for line in benchmarks:
        line = line.split('|')
        nd = int(line[0].strip())
        nk1 = int(line[1].strip())
        nk2 = int(line[2].strip())
        actual = float(line[4].strip())
        cost1, cost2 = predictions(nd, nk1, nk2)
        predicted = cost1 / cost2
        correct = ((predicted >= 1 and actual >= 1) or
                   (predicted < 1 and actual < 1))
        print "%6d | %8d | %8d | %9.2f | %6.2f | %s" % (
            nd, nd/nk1, nd/nk2,  predicted, actual, correct)
        count += 1
        if correct:
            n_correct += 1

        if cost1 < cost2:
            # I picked algorithm1, so no net loss or gain
            gain += 1.0
        else:
            # I picked algorith2, so note difference in performance
            gain += actual

    print "-" * 79
    print "%% correct: %0.1f" % (n_correct * 100.0 / count)
    print "%% performance gain: %0.1f" % ((gain / count - 1.0) * 100.0)

if __name__ == '__main__':
    #do_benchmark('benchmark.db', 10000, 1000, 1000)
    #do_benchmarks('/dev/shm/benchmark.db')
    rerun_predictions('benchmarks.txt')



import unittest

_marker = object()


class PathIndexTests(unittest.TestCase):
    """ Test PathIndex objects
    """
    def _getTargetClass(self):
        from zerodbext.catalog.indexes.path import CatalogPathIndex
        return CatalogPathIndex

    def _makeOne(self, values=None, discriminator=_marker):
        if values is None:
            values = {}
        def _discriminator(obj, default):
            if obj is _marker:
                return default
            return obj.path
        if discriminator is _marker:
            discriminator = _discriminator
        index = self._getTargetClass()(discriminator)
        for doc_id, path in values.items():
            index.index_doc(doc_id, path)
        return index

    def test_class_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyClass
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyClass(ICatalogIndex, self._getTargetClass())

    def test_instance_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyObject
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyObject(ICatalogIndex, self._makeOne())

    def test_ctor_callback_discriminator(self):
        def _discriminator(obj, default):
            """ """
        index = self._makeOne(discriminator=_discriminator)
        self.assertTrue(index.discriminator is _discriminator)

    def test_ctor_string_discriminator(self):
        index = self._makeOne(discriminator='abc')
        self.assertEqual(index.discriminator, 'abc')

    def test_ctor_bad_discriminator(self):
        self.assertRaises(ValueError, self._makeOne, discriminator=object())

    def test_empty_index(self):
        index = self._makeOne({})
        self.assertEqual(index.numObjects() ,0)
        self.assertEqual(index._depth, 0)
        self.assertEqual(index.getEntryForObject(1234), None)
        index.unindex_doc(1234) # nothrow
        result = index.apply({"suxpath":"xxx"})
        self.assertEqual(list(result), [])

    def test_nonempty_index(self):
        index = self._makeOne(VALUES)
        self.assertEqual(index.numObjects(), 18)
        self.assertEqual(index.getEntryForObject(1), '/aa/aa/aa/1.html')

    def test_clear(self):
        index = self._makeOne(VALUES)
        index.clear()
        self.assertEqual(index.numObjects(), 0)
        self.assertEqual(index.getEntryForObject(1), None)
        self.assertEqual(index._depth, 0)

    def test_insertEntry_new_component_new_level(self):
        index = self._makeOne()
        index.insertEntry('aaa', 1, 1)
        self.assertEqual(len(index._index), 1)
        self.assertEqual(len(index._index['aaa']), 1)
        self.assertEqual(list(index._index['aaa'][1]), [1])
        self.assertEqual(index._depth, 1)

    def test_insertEntry_existing_component_new_level(self):
        index = self._makeOne()
        index._index['aaa'] = {2: FakeTreeSet()}
        index._depth = 2
        index.insertEntry('aaa', 1, 1)
        self.assertEqual(len(index._index), 1)
        self.assertEqual(len(index._index['aaa']), 2)
        self.assertEqual(list(index._index['aaa'][1]), [1])
        self.assertEqual(index._depth, 2)

    def test_insertEntry_existing_component_exsiting_level(self):
        index = self._makeOne()
        fts = FakeTreeSet()
        index._index['aaa'] = {2: fts}
        index._depth = 2
        index.insertEntry('aaa', 1, 2)
        self.assertEqual(len(index._index), 1)
        self.assertEqual(len(index._index['aaa']), 1)
        self.assertTrue(index._index['aaa'][2] is fts)
        self.assertEqual(list(index._index['aaa'][2]), [1])
        self.assertEqual(index._depth, 2)

    def test_index_doc_callback_returns_nondefault(self):
        def callback(ob, default):
            return ob
        index = self._makeOne(discriminator=callback)
        index.index_doc(1, '/a/b/c')
        self.assertEqual(len(index._index), 3)
        self.assertEqual(list(index._index['a'][0]), [1])
        self.assertEqual(list(index._index['b'][1]), [1])
        self.assertEqual(list(index._index['c'][2]), [1])
        self.assertEqual(index._depth, 2)

    def test_index_doc_string_discrim(self):
        index = self._makeOne(discriminator='abc')
        class Dummy:
            abc = '/a/b/c'
        dummy = Dummy()
        index.index_doc(1, dummy)
        self.assertEqual(len(index._index), 3)
        self.assertEqual(list(index._index['a'][0]), [1])
        self.assertEqual(list(index._index['b'][1]), [1])
        self.assertEqual(list(index._index['c'][2]), [1])
        self.assertEqual(index._depth, 2)

    def test_index_doc_string_discrim_tuple_value(self):
        index = self._makeOne(discriminator='abc')
        class Dummy:
            abc = ('', 'a', 'b', 'c')
        dummy = Dummy()
        index.index_doc(1, dummy)
        self.assertEqual(len(index._index), 3)
        self.assertEqual(list(index._index['a'][0]), [1])
        self.assertEqual(list(index._index['b'][1]), [1])
        self.assertEqual(list(index._index['c'][2]), [1])
        self.assertEqual(index._depth, 2)

    def test_index_doc_string_discrim_list_value(self):
        index = self._makeOne(discriminator='abc')
        class Dummy:
            abc = ['', 'a', 'b', 'c']
        dummy = Dummy()
        index.index_doc(1, dummy)
        self.assertEqual(len(index._index), 3)
        self.assertEqual(list(index._index['a'][0]), [1])
        self.assertEqual(list(index._index['b'][1]), [1])
        self.assertEqual(list(index._index['c'][2]), [1])
        self.assertEqual(index._depth, 2)

    def test_index_doc_missing_value_unindexes(self):
        index = self._makeOne(discriminator='abc')
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = '/a/b/c'
        index.index_doc(1, dummy)
        del dummy.abc
        index.index_doc(1, dummy)
        self.assertFalse('a' in index._index)
        self.assertFalse('b' in index._index)
        self.assertFalse('c' in index._index)
        self.assertEqual(index._depth, 2)

    def test_index_doc_persistent_value_raises(self):
        from persistent import Persistent
        index = self._makeOne(discriminator='abc')
        class Dummy:
            pass
        dummy = Dummy()
        dummy.abc = Persistent()
        self.assertRaises(ValueError, index.index_doc, 1, dummy)

    def test_unindex_doc(self):
        index = self._makeOne(VALUES)

        for doc_id in VALUES.keys():
            index.unindex_doc(doc_id)

        self.assertEqual(index.numObjects(), 0)
        self.assertEqual(len(index._index), 0)
        self.assertEqual(len(index._unindex), 0)

    def test_reindex_doesnt_dupe(self):
        index = self._makeOne()
        o = Dummy('/foo/bar')
        index.index_doc(1, o)
        self.assertEqual(index.numObjects(), 1)
        index.index_doc(1, o)
        self.assertEqual(index.numObjects(), 1)

    def test_unindex_nomatch_doesnt_raise(self):
        index = self._makeOne()
        # this should not raise an error
        index.unindex_doc(-1)

        # nor should this
        index._unindex[1] = "/broken/thing"
        index.unindex_doc(1)

    def test_searches_against_root_plain_string(self):
        index = self._makeOne(VALUES)
        expected = list(range(1,19))

        results = list(index.apply('/').keys())
        self.assertEqual(results, expected)

    def test_searches_against_root_tuple(self):
        index = self._makeOne(VALUES)
        expected = list(range(1,19))

        results = list(index.apply(('',)).keys())
        self.assertEqual(results, expected)

    def test_searches_against_root_list(self):
        index = self._makeOne(VALUES)
        expected = list(range(1,19))

        results = list(index.apply(['']).keys())
        self.assertEqual(results, expected)

    def test_searches_against_root_wo_level(self):
        index = self._makeOne(VALUES)
        expected = list(range(1,19))

        results = list(index.apply({'query': '/'}).keys())
        self.assertEqual(results, expected)

    def test_searches_against_root_w_level(self):
        index = self._makeOne(VALUES)
        comp = "/"
        level = 0
        expected = list(range(1,19))

        results = list(index.apply({'query': '/', 'level': 0}).keys())
        self.assertEqual(results, expected)

    def test_root_aa_wo_level(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,7,8,9]

        results = list(index.apply({'query': '/aa'}).keys())
        self.assertEqual(results, expected)

    def test_root_aa_tuple(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,7,8,9]

        results = list(index.apply(('/aa', '/cc')).keys())
        self.assertEqual(results, expected)

    def test_root_aa_list(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,7,8,9]

        results = list(index.apply(['/aa', '/cc']).keys())
        self.assertEqual(results, expected)

    def test_aa_explicit_level_0(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,7,8,9]

        results = list(index.apply({'query': 'aa', 'level': 0}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('aa', 0)]}).keys())
        self.assertEqual(results, expected)

    def test_aa_level_1(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,10,11,12]

        results = list(index.apply({'query': 'aa', 'level': 1}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('aa', 1)]}).keys())
        self.assertEqual(results, expected)

    def test_aa_level_4(self):
        index = self._makeOne(VALUES)
        expected = []

        results = list(index.apply({'query': 'aa', 'level': 4}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('aa', 4)]}).keys())
        self.assertEqual(results, expected)

    def test_bb_level_0(self):
        index = self._makeOne(VALUES)
        expected = [10,11,12,13,14,15,16,17,18]

        results = list(index.apply({'query': 'bb', 'level': 0}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('bb', 0)]}).keys())
        self.assertEqual(results, expected)

    def test_bb_level_1(self):
        index = self._makeOne(VALUES)
        expected = [4,5,6,13,14,15]

        results = list(index.apply({'query': 'bb', 'level': 1}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('bb', 1)]}).keys())
        self.assertEqual(results, expected)

    def test_bb_cc_level_0(self):
        index = self._makeOne(VALUES)
        expected = [16,17,18]

        results = list(index.apply({'query': 'bb/cc', 'level': 0}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('bb/cc', 0)]}).keys())
        self.assertEqual(results, expected)

    def test_bb_cc_level_1(self):
        index = self._makeOne(VALUES)
        expected = [6,15]

        results = list(index.apply({'query': 'bb/cc', 'level': 1}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('bb/cc', 1)]}).keys())
        self.assertEqual(results, expected)

    def test_bb_aa_level_0(self):
        index = self._makeOne(VALUES)
        expected = [10,11,12]

        results = list(index.apply({'query': 'bb/aa', 'level': 0}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('bb/aa', 0)]}).keys())
        self.assertEqual(results, expected)

    def test_bb_aa_level_1(self):
        index = self._makeOne(VALUES)
        expected = [4,13]

        results = list(index.apply({'query': 'bb/aa', 'level': 1}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('bb/aa', 1)]}).keys())
        self.assertEqual(results, expected)

    def test_aa_cc_level_neg1(self):
        index = self._makeOne(VALUES)
        expected = [3,7,8,9,12]

        results = list(index.apply({'query': 'aa/cc', 'level': -1}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('aa/cc', -1)]}).keys())
        self.assertEqual(results, expected)

    def test_bb_bb_level_neg1(self):
        index = self._makeOne(VALUES)
        expected = [5,13,14,15]

        results = list(index.apply({'query': 'bb/bb', 'level': -1}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('bb/bb', -1)]}).keys())
        self.assertEqual(results, expected)

    def test_18_html_level_3(self):
        index = self._makeOne(VALUES)
        expected = [18]

        results = list(index.apply({'query': '18.html', 'level': 3}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('18.html', 3)]}).keys())
        self.assertEqual(results, expected)

    def test_18_html_level_neg1(self):
        index = self._makeOne(VALUES)
        expected = [18]

        results = list(index.apply({'query': '18.html', 'level': -1}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('18.html', -1)]}).keys())
        self.assertEqual(results, expected)

    def test_cc_18_html_level_2(self):
        index = self._makeOne(VALUES)
        expected = [18]

        results = list(index.apply({'query': 'cc/18.html', 'level': 2}).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('cc/18.html', 2)]}).keys())
        self.assertEqual(results, expected)

    def test_cc_18_html_level_neg1(self):
        index = self._makeOne(VALUES)
        expected = [18]

        results = list(index.apply({'query': 'cc/18.html', 'level': -1}
                                  ).keys())
        self.assertEqual(results, expected)

        results = list(index.apply({'query': [('cc/18.html', -1)]}).keys())
        self.assertEqual(results, expected)

    def test_aa_bb_implicit_OR_level_1(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,10,11,12,13,14,15]
        results = list(index.apply({'query': ['aa', 'bb'],
                                    'level': 1,
                                   }).keys())
        self.assertEqual(results, expected)

    def test_aa_bb_explicit_OR_level_1(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,10,11,12,13,14,15]
        results = list(index.apply({'query': ['aa', 'bb'],
                                    'level': 1,
                                    'operator': 'or',
                                   }).keys())
        self.assertEqual(results, expected)

    def test_aa_bb_xx_implicit_OR_level_1(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,10,11,12,13,14,15]
        results = list(index.apply({'query': ['aa', 'bb', 'xx'],
                                    'level': 1,
                                   }).keys())
        self.assertEqual(results, expected)

    def test_aa_bb_xx_explicit_OR_level_1(self):
        index = self._makeOne(VALUES)
        expected = [1,2,3,4,5,6,10,11,12,13,14,15]
        results = list(index.apply({'query': ['aa', 'bb', 'xx'],
                                    'level': 1,
                                    'operator': 'or',
                                   }).keys())
        self.assertEqual(results, expected)

    def test_cc_level_1_cc_level_2_implicit_OR(self):
        index = self._makeOne(VALUES)
        expected = [3,6,7,8,9,12,15,16,17,18]
        results = list(index.apply({'query': [('cc', 1), ('cc', 2)],
                                   }).keys())
        self.assertEqual(results, expected)

    def test_cc_level_1_cc_level_2_explicit_OR(self):
        index = self._makeOne(VALUES)
        expected = [3,6,7,8,9,12,15,16,17,18]
        results = list(index.apply({'query': [('cc', 1), ('cc', 2)],
                                    'operator': 'or',
                                   }).keys())
        self.assertEqual(results, expected)

    def test_aa_AND_bb_level_1(self):
        index = self._makeOne(VALUES)
        expected = []
        results = list(index.apply({'query': ['aa', 'bb'],
                                    'level': 1,
                                    'operator': 'and',
                                   }).keys())
        self.assertEqual(results, expected)

    def test_aa_level_0_AND_bb_level_1(self):
        index = self._makeOne(VALUES)
        expected = [4, 5, 6]
        results = list(index.apply({'query': [('aa', 0), ('bb', 1)],
                                    'operator': 'and',
                                   }).keys())
        self.assertEqual(results, expected)

    def test_aa_level_0_AND_cc_level_2(self):
        index = self._makeOne(VALUES)
        expected = [3, 6, 9]
        results = list(index.apply({'query': [('aa', 0), ('cc', 2)],
                                    'operator': 'and',
                                   }).keys())
        self.assertEqual(results, expected)

    def test_query_path_included_with_childrent(self):
        index = self._makeOne()
        index.index_doc(1, Dummy("/ff"))
        index.index_doc(2, Dummy("/ff/gg"))
        index.index_doc(3, Dummy("/ff/gg/3.html"))
        index.index_doc(4, Dummy("/ff/gg/4.html"))
        result = list(index.apply({'query':'/ff/gg'}).keys())
        self.assertEqual(result, [2, 3, 4])

    def test_applyEq(self):
        index = self._makeOne(VALUES)
        expected = list(range(1,19))

        results = list(index.applyEq('/').keys())
        self.assertEqual(results, expected)

    def test_applyNotEq(self):
        index = self._makeOne(VALUES)
        expected = list(range(1,10))

        results = list(index.applyNotEq('/bb').keys())
        self.assertEqual(results, expected)

    def test_docids(self):
        index = self._makeOne(VALUES)
        self.assertEqual(set(index.docids()), set(range(1, 19)))

    def test_unindex_doc_removes_from_docids(self):
        index = self._makeOne(VALUES)
        index.index_doc(20, _marker)
        index.unindex_doc(20)
        self.assertFalse(20 in index.docids())

    def test_index_doc_then_missing_value(self):
        index = self._makeOne(VALUES)
        self.assertEqual(set([3]), set(index.applyEq('/aa/aa/cc/3.html')))
        self.assertTrue(3 in index.docids())
        index.index_doc(3, _marker)
        self.assertEqual(set(), set(index.applyEq('/aa/aa/cc/3.html')))
        self.assertTrue(3 in index.docids())

    def test_index_doc_missing_value_then_with_value(self):
        index = self._makeOne()
        index.index_doc(20, _marker)
        self.assertEqual(set(), set(index.applyEq('/cmr')))
        self.assertTrue(20 in index.docids())
        index.index_doc(20, Dummy('/cmr/ljb'))
        self.assertEqual(set([20]), set(index.applyEq('/cmr')))
        self.assertTrue(20 in index.docids())

    def test_index_doc_missing_value_then_unindex(self):
        index = self._makeOne()
        index.index_doc(20, _marker)
        self.assertEqual(set(), set(index.applyEq('/cmr')))
        self.assertTrue(20 in index.docids())
        index.unindex_doc(20)
        self.assertEqual(set(), set(index.applyEq('/cmr')))
        self.assertFalse(20 in index.docids())

    def test_docids_with_indexed_and_not_indexed(self):
        index = self._makeOne()
        index.index_doc(1, Dummy('/cmr/ljb'))
        index.index_doc(2, _marker)
        self.assertEqual(set([1, 2]), set(index.docids()))


class FakeTreeSet(set):
    def insert(self, thing):
        self.add(thing)


class Dummy:

    def __init__( self, path):
        self.path = path


VALUES = {
    1 : Dummy("/aa/aa/aa/1.html"),
    2 : Dummy("/aa/aa/bb/2.html"),
    3 : Dummy("/aa/aa/cc/3.html"),
    4 : Dummy("/aa/bb/aa/4.html"),
    5 : Dummy("/aa/bb/bb/5.html"),
    6 : Dummy("/aa/bb/cc/6.html"),
    7 : Dummy("/aa/cc/aa/7.html"),
    8 : Dummy("/aa/cc/bb/8.html"),
    9 : Dummy("/aa/cc/cc/9.html"),
    10 : Dummy("/bb/aa/aa/10.html"),
    11 : Dummy("/bb/aa/bb/11.html"),
    12 : Dummy("/bb/aa/cc/12.html"),
    13 : Dummy("/bb/bb/aa/13.html"),
    14 : Dummy("/bb/bb/bb/14.html"),
    15 : Dummy("/bb/bb/cc/15.html"),
    16 : Dummy("/bb/cc/aa/16.html"),
    17 : Dummy("/bb/cc/bb/17.html"),
    18 : Dummy("/bb/cc/cc/18.html")
}



import unittest

_marker = object()


class TestCatalogFieldIndex(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.indexes.field import CatalogFieldIndex
        return CatalogFieldIndex

    def _makeOne(self, discriminator=_marker):
        def _discriminator(obj, default):
            if obj is _marker:
                return default
            return obj
        if discriminator is _marker:
            discriminator = _discriminator
        return self._getTargetClass()(discriminator)

    def _populateIndex(self, index):
        index.index_doc(5, 1)  # docid, obj
        index.index_doc(2, 2)
        index.index_doc(1, 3)
        index.index_doc(3, 4)
        index.index_doc(4, 5)
        index.index_doc(8, 6)
        index.index_doc(9, 7)
        index.index_doc(7, 8)
        index.index_doc(6, 9)
        index.index_doc(11, 10)
        index.index_doc(10, 11)

    def test_class_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyClass
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyClass(ICatalogIndex, self._getTargetClass())

    def test_instance_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyObject
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyObject(ICatalogIndex, self._makeOne())

    def test_ctor_callback_discriminator(self):
        def _discriminator(obj, default):
            """ """
        index = self._makeOne(_discriminator)
        self.assertTrue(index.discriminator is _discriminator)

    def test_ctor_string_discriminator(self):
        index = self._makeOne('abc')
        self.assertEqual(index.discriminator, 'abc')

    def test_ctor_bad_discriminator(self):
        self.assertRaises(ValueError, self._makeOne, object())

    def test_reindex_doc_doesnt_unindex(self):
        index = self._makeOne()
        index.index_doc(5, 1)
        index.unindex_doc = lambda *args, **kw: 1 / 0
        index.reindex_doc(5, 1)

    def test_reindex_doc_w_existing_docid_same_value(self):
        index = self._makeOne()
        index.index_doc(5, 1)
        self.assertEqual(index.documentCount(), 1)
        self.assertEqual(index._rev_index[5], 1)
        index.reindex_doc(5, 1)
        self.assertEqual(index.documentCount(), 1)
        self.assertEqual(index._rev_index[5], 1)

    def test_reindex_doc_w_existing_docid_different_value(self):
        index = self._makeOne()
        index.index_doc(5, 1)
        self.assertEqual(index.documentCount(), 1)
        index.reindex_doc(5, 2)
        self.assertEqual(index.documentCount(), 1)
        self.assertEqual(index._rev_index[5], 2)

    def test_reindex_doc_w_new_docid(self):
        index = self._makeOne()
        index.index_doc(5, 1)
        self.assertEqual(index.documentCount(), 1)
        index.reindex_doc(6, 2)
        self.assertEqual(index.documentCount(), 2)
        self.assertEqual(index._rev_index[5], 1)
        self.assertEqual(index._rev_index[6], 2)

    def test_unindex_doc_nonesuch(self):
        index = self._makeOne()
        index.index_doc(5, 1)
        index.unindex_doc(6)
        self.assertEqual(index.documentCount(), 1)

    def test_unindex_doc_no_other_docids_for_value(self):
        index = self._makeOne()
        index.index_doc(5, 1)
        index.unindex_doc(5)
        self.assertEqual(index.documentCount(), 0)
        self.assertFalse(5 in index._rev_index)
        self.assertFalse(1 in index._fwd_index)

    def test_unindex_doc_w_other_docids_for_value(self):
        index = self._makeOne()
        index.index_doc(5, 1)
        index.index_doc(6, 1)
        index.unindex_doc(5)
        self.assertEqual(index.documentCount(), 1)
        self.assertFalse(5 in index._rev_index)
        self.assertTrue(1 in index._fwd_index)

    def test_sort_no_docids(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet()
        result = index.sort(c1)
        self.assertEqual(list(result), [])

    def test_sort_no_docs(self):
        index = self._makeOne()
        from BTrees.IFBTree import IFSet
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1)
        self.assertEqual(list(result), [])

    def test_sort_bad_limit(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        self.assertRaises(ValueError, index.sort, c1, limit=0)

    def test_sort_bad_sort_type(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        self.assertRaises(ValueError, index.sort, c1, sort_type='nonesuch')

    def test_sort_bad_sort_type_reverse_fwscan(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import FWSCAN
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        self.assertRaises(ValueError,
                          index.sort, c1, reverse=True, sort_type=FWSCAN)

    def test_sort_bad_sort_type_reverse(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        self.assertRaises(ValueError,
                          index.sort, c1, reverse=True, sort_type='nonesuch')

    def test_sort_noforce_no_limit(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1)
        self.assertEqual(list(result), [5, 2, 1, 3, 4])

    def test_sort_noforce_timsort(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        # Add enough items to make 'fwscan_wins' return False.
        for i in range(100, 10000):
            index.index_doc(i, i)
        c1 = IFSet([1, 2, 3, 4, 5])
        # No 'limit' means timsort
        result = index.sort(c1)
        self.assertEqual(list(result), [5, 2, 1, 3, 4])

    def test_sort_noforce_nbest(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        # Add enough items to make 'fwscan_wins' return False.
        for i in range(100, 10000):
            index.index_doc(i, i)
        c1 = IFSet([1, 2, 3, 4, 5])
        # Reasonable 'limit' means nbest
        result = index.sort(c1, limit=3)
        self.assertEqual(list(result), [5, 2, 1])

    def test_sort_noforce_w_limit(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, limit=3)
        self.assertEqual(list(result), [5, 2, 1])

    def test_sort_noforce_reverse_no_limit(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, reverse=True)
        self.assertEqual(list(result), [4, 3, 1, 2, 5])

    def test_sort_noforce_reverse_w_limit(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, reverse=True, limit=3)
        self.assertEqual(list(result), [4, 3, 1])

    def test_sort_noforce_reverse_w_limit_timsort(self):
        from BTrees.IFBTree import IFSet
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        # Add enough items to make 'fwscan_wins' return False.
        for i in range(100, 10000):
            index.index_doc(i, i)
            c1.insert(i)
        result = index.sort(c1, reverse=True, limit=301) # waaa
        self.assertEqual(list(result), list(range(9999, 9698, -1)))

    def test_sort_force_fwscan_no_limit(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import FWSCAN
        index = self._makeOne()
        index.force_scan = True
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, sort_type=FWSCAN)
        self.assertEqual(list(result), [5, 2, 1, 3, 4])

    def test_sort_force_fwscan_missing_docid(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import FWSCAN
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5, 99])
        result = index.sort(c1, sort_type=FWSCAN)
        r = list(result)
        self.assertEqual(r, [5, 2, 1, 3, 4]) # 99 not present

    def test_sort_force_fwscan_w_limit(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import FWSCAN
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, limit=3, sort_type=FWSCAN)
        self.assertEqual(list(result), [5, 2, 1])

    def test_sort_force_timsort_no_limit(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import TIMSORT
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, sort_type=TIMSORT)
        self.assertEqual(list(result), [5, 2, 1, 3, 4])

    def test_sort_force_timsort_missing_docid(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import TIMSORT
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5, 99])
        result = index.sort(c1, sort_type=TIMSORT)
        r = list(result)
        self.assertEqual(r, [5, 2, 1, 3, 4]) # 99 not present

    def test_sort_force_timsort_w_limit(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import TIMSORT
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, limit=3, sort_type=TIMSORT)
        self.assertEqual(list(result), [5, 2, 1])

    def test_sort_force_timsort_reverse_nolimit(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import TIMSORT
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, reverse=True, sort_type=TIMSORT)
        self.assertEqual(list(result), [4, 3, 1, 2, 5])

    def test_sort_force_timsort_reverse_missing_docid(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import TIMSORT
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5, 99])
        result = index.sort(c1, reverse=True, sort_type=TIMSORT)
        r = list(result)
        self.assertEqual(r, [4, 3, 1, 2, 5]) # 99 not present

    def test_sort_force_timsort_reverse_withlimit(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import TIMSORT
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, reverse=True, limit=3, sort_type=TIMSORT)
        self.assertEqual(list(result), [4, 3, 1])

    def test_sort_force_nbest_no_limit_raises(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import NBEST
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        self.assertRaises(ValueError, index.sort, c1, sort_type=NBEST)

    def test_sort_force_nbest(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import NBEST
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, limit=3, sort_type=NBEST)
        self.assertEqual(list(result), [5, 2, 1])

    def test_sort_force_nbest_reverse_no_limit_raises(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import NBEST
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        self.assertRaises(ValueError,
                          index.sort, c1, reverse=True, sort_type=NBEST)

    def test_sort_force_nbest_reverse(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import NBEST
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5])
        result = index.sort(c1, reverse=True, limit=3, sort_type=NBEST)
        self.assertEqual(list(result), [4, 3, 1])

    def test_sort_force_nbest_missing_docid(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import NBEST
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5, 99])
        result = index.sort(c1, limit=3, sort_type=NBEST)
        self.assertEqual(list(result), [5, 2, 1])

    def test_sort_force_nbest_reverse_missing_docid(self):
        from BTrees.IFBTree import IFSet
        from zerodbext.catalog.indexes.field import NBEST
        index = self._makeOne()
        self._populateIndex(index)
        c1 = IFSet([1, 2, 3, 4, 5, 99])
        result = index.sort(c1, reverse=True, limit=3, sort_type=NBEST)
        self.assertEqual(list(result), [4, 3, 1])

    def test_search_single_range_querymember_or(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        from zerodbext.catalog import RangeValue
        result = index.search([RangeValue(1,1)])
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_search_double_range_querymember_or(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        from zerodbext.catalog import RangeValue
        result = index.search([RangeValue(1,1), RangeValue(1,2)])
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_search_double_range_querymember_and(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        from zerodbext.catalog import RangeValue
        result = index.search([RangeValue(1,1), RangeValue(1,2)], 'and')
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_search_single_int_querymember_or(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.search([1])
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_search_double_int_querymember_or(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.search([1, 2])
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_search_double_int_querymember_and(self):
        # this is a nonsensical query
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.search([1, 2], 'and')
        result = sorted(list(result))
        self.assertEqual(result, [])

    def test_apply_dict_single_range(self):
        from zerodbext.catalog import RangeValue
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query': RangeValue(1,2)})
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_apply_dict_operator_or_with_ranges(self):
        from zerodbext.catalog import RangeValue
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':[RangeValue(1,1), RangeValue(1,2)],
                              'operator':'or'})
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_apply_dict_operator_and_with_ranges_and(self):
        from zerodbext.catalog import RangeValue
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':[RangeValue(1,1), RangeValue(1,2)],
                              'operator':'and'})
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_apply_dict_operator_and_with_ranges_or(self):
        from zerodbext.catalog import RangeValue
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':[RangeValue(1,1), RangeValue(1,2)],
                              'operator':'or'})
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_apply_dict_operator_or_with_single_int(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':1})
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_apply_dict_operator_or_with_list_of_ints_or(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':[1,2]})
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_apply_dict_operator_or_with_list_of_ints_and(self):
        # nonsensical query
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':[1, 2], 'operator':'and'})
        result = sorted(list(result))
        self.assertEqual(result, [])

    def test_apply_dict_operator_or_with_int_and_range_or(self):
        from zerodbext.catalog import RangeValue
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':[1, RangeValue(1,2)], 'operator':'or'})
        result = sorted(list(result))
        self.assertEqual(result, [2,5,50])

    def test_apply_dict_operator_or_with_int_and_range_and(self):
        from zerodbext.catalog import RangeValue
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply({'query':[1, RangeValue(1,2)], 'operator':'and'})
        result = sorted(list(result))
        self.assertEqual(result, [5,50])

    def test_apply_nondict_2tuple(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply((1,2))
        result = sorted(list(result))
        self.assertEqual(result, [2, 5,50])

    def test_apply_nondict_int(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply(1)
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_apply_list(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.apply([1,2])
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_applyEq(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyEq(1)
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_applyNotEq(self):
        index = self._makeOne()
        self._populateIndex(index)
        result = index.applyNotEq(1)
        result = sorted(list(result))
        self.assertEqual(result, [1,2,3,4,6,7,8,9,10,11])

    def test_applyNotEq_returns_unindexed_docs(self):
        def discriminator(obj, default):
            if isinstance(obj, int):
                return obj
            return default

        index = self._makeOne(discriminator)
        self._populateIndex(index)
        index.index_doc(50, 1)
        index.index_doc(51, '1')
        result = index.applyNotEq(1)
        result = sorted(list(result))
        self.assertEqual(result, [1,2,3,4,6,7,8,9,10,11,51])

    def test_applyNotEq_nothing_indexed(self):
        index = self._makeOne('foo')
        self._populateIndex(index)
        result = index.applyNotEq(1)
        result = sorted(list(result))
        self.assertEqual(result, [1,2,3,4,5,6,7,8,9,10,11])

    def test_applyGe(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyGe(10)
        result = sorted(list(result))
        self.assertEqual(result, [10, 11])

    def test_applyGt(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyGt(10)
        result = sorted(list(result))
        self.assertEqual(result, [10])

    def test_applyLe(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyLe(2)
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_applyLt(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyLt(2)
        result = sorted(list(result))
        self.assertEqual(result, [5, 50])

    def test_applyAny(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyAny([1, 2, 60])
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 50])

    def test_applyAny_noresults(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyAny([60, 70])
        result = sorted(list(result))
        self.assertEqual(result, [])

    def test_applyNotAny(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(50, 1)
        result = index.applyNotAny([1, 2, 60])
        result = sorted(list(result))
        self.assertEqual(result, [1, 3, 4, 6, 7, 8, 9, 10, 11])

    def test_applyInRange_inclusive_inclusive(self):
        index = self._makeOne()
        self._populateIndex(index)
        result = index.applyInRange(3, 7)
        result = sorted(list(result))
        self.assertEqual(result, [1, 3, 4, 8, 9])

    def test_applyInRange_inclusive_exclusive(self):
        index = self._makeOne()
        self._populateIndex(index)
        result = index.applyInRange(3, 7, excludemax=True)
        result = sorted(list(result))
        self.assertEqual(result, [1, 3, 4, 8])

    def test_applyInRange_exclusive_inclusive(self):
        index = self._makeOne()
        self._populateIndex(index)
        result = index.applyInRange(3, 7, excludemin=True)
        result = sorted(list(result))
        self.assertEqual(result, [3, 4, 8, 9])

    def test_applyInRange_exclusive_exclusive(self):
        index = self._makeOne()
        self._populateIndex(index)
        result = index.applyInRange(3, 7, excludemin=True, excludemax=True)
        result = sorted(list(result))
        self.assertEqual(result, [3, 4, 8])

    def test_applyNotInRange(self):
        index = self._makeOne()
        self._populateIndex(index)
        result = index.applyNotInRange(3, 7)
        result = sorted(list(result))
        self.assertEqual(result, [2, 5, 6, 7, 10, 11])

    def test_docids(self):
        index = self._makeOne()
        self._populateIndex(index)
        self.assertEqual(
            set(index.docids()),
            set((1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)))

    def test_unindex_doc_removes_from_docids(self):
        index = self._makeOne()
        index.index_doc(20, _marker)
        index.unindex_doc(20)
        self.assertFalse(20 in index.docids())

    def test_index_doc_then_missing_value(self):
        index = self._makeOne()
        self._populateIndex(index)
        self.assertEqual(set([3]), set(index.applyEq(4)))
        self.assertTrue(3 in index.docids())
        index.index_doc(3, _marker)
        self.assertEqual(set(), set(index.applyEq(4)))
        self.assertTrue(3 in index.docids())

    def test_index_doc_missing_value_then_with_value(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(3, _marker)
        self.assertEqual(set(), set(index.applyEq(4)))
        self.assertTrue(3 in index.docids())
        index.index_doc(3, 42)
        self.assertEqual(set([3]), set(index.applyEq(42)))
        self.assertTrue(3 in index.docids())

    def test_index_doc_missing_value_then_unindex(self):
        index = self._makeOne()
        self._populateIndex(index)
        index.index_doc(3, _marker)
        self.assertEqual(set(), set(index.applyEq(4)))
        self.assertTrue(3 in index.docids())
        index.unindex_doc(3)
        self.assertEqual(set(), set(index.applyEq(4)))
        self.assertFalse(3 in index.docids())

    def test_docids_with_indexed_and_not_indexed(self):
        index = self._makeOne()
        index.index_doc(1, 1)
        index.index_doc(2, _marker)
        self.assertEqual(set([1, 2]), set(index.docids()))


class Test_fwscan_wins(unittest.TestCase):

    def _callFUT(self, limit, rlen, numdocs):
        from zerodbext.catalog.indexes.field import fwscan_wins
        return fwscan_wins(limit, rlen, numdocs)

    def test_numdocs_zero_raises(self):
        self.assertRaises(ZeroDivisionError, self._callFUT, None, 100, 0)

    def test_no_limit_rlen_gt_one_fourth_numdocs(self):
        self.assertTrue(self._callFUT(None, 25, 100))
        self.assertTrue(self._callFUT(None, 26, 100))
        self.assertTrue(self._callFUT(None, 50, 100))
        self.assertTrue(self._callFUT(None, 75, 100))
        self.assertTrue(self._callFUT(None, 99, 100))

    def test_no_limit_rlen_ge_one_fourth_numdocs(self):
        self.assertFalse(self._callFUT(None, 24, 100))
        self.assertFalse(self._callFUT(None, 23, 100))
        self.assertFalse(self._callFUT(None, 15, 100))
        self.assertFalse(self._callFUT(None, 10, 100))
        self.assertFalse(self._callFUT(None, 5, 100))
        self.assertFalse(self._callFUT(None, 2, 100))
        self.assertFalse(self._callFUT(None, 1, 100))

    def test_w_limit_docratio_ge_512_lt_1024_over_div(self):
        self.assertTrue(self._callFUT(1, 512, 65536))
        self.assertTrue(self._callFUT(2, 512, 65536))
        self.assertTrue(self._callFUT(3, 512, 65536))
        self.assertTrue(self._callFUT(4, 512, 65536))
        self.assertFalse(self._callFUT(5, 512, 65536))
        self.assertTrue(self._callFUT(1, 1023, 65536))
        self.assertTrue(self._callFUT(2, 1023, 65536))
        self.assertTrue(self._callFUT(3, 1023, 65536))
        self.assertTrue(self._callFUT(4, 1023, 65536))
        self.assertFalse(self._callFUT(5, 1023, 65536))

    def test_w_limit_docratio_ge_1024_lt_2048_over_div(self):
        self.assertTrue(self._callFUT(1, 1024, 65536))
        self.assertTrue(self._callFUT(2, 1024, 65536))
        self.assertTrue(self._callFUT(31, 1024, 65536))
        self.assertTrue(self._callFUT(32, 1024, 65536))
        self.assertFalse(self._callFUT(33, 1024, 65536))
        self.assertTrue(self._callFUT(1, 2047, 65536))
        self.assertTrue(self._callFUT(2, 2047, 65536))
        self.assertTrue(self._callFUT(31, 2047, 65536))
        self.assertTrue(self._callFUT(32, 2047, 65536))
        self.assertFalse(self._callFUT(33, 2047, 65536))

    def test_w_limit_docratio_ge_2048_lt_4096_over_div(self):
        self.assertTrue(self._callFUT(1, 2048, 65536))
        self.assertTrue(self._callFUT(2, 2048, 65536))
        self.assertTrue(self._callFUT(127, 2048, 65536))
        self.assertTrue(self._callFUT(128, 2048, 65536))
        self.assertFalse(self._callFUT(129, 2048, 65536))
        self.assertTrue(self._callFUT(1, 4095, 65536))
        self.assertTrue(self._callFUT(2, 4095, 65536))
        self.assertTrue(self._callFUT(127, 4095, 65536))
        self.assertTrue(self._callFUT(128, 4095, 65536))
        self.assertFalse(self._callFUT(129, 4095, 65536))

    def test_w_limit_docratio_ge_4096_lt_8092_over_div(self):
        self.assertTrue(self._callFUT(1, 4096, 65536))
        self.assertTrue(self._callFUT(2, 4096, 65536))
        self.assertTrue(self._callFUT(511, 4096, 65536))
        self.assertTrue(self._callFUT(512, 4096, 65536))
        self.assertFalse(self._callFUT(513, 4096, 65536))
        self.assertTrue(self._callFUT(1, 8091, 65536))
        self.assertTrue(self._callFUT(2, 8091, 65536))
        self.assertTrue(self._callFUT(511, 8091, 65536))
        self.assertTrue(self._callFUT(512, 8091, 65536))
        self.assertFalse(self._callFUT(513, 8091, 65536))

    def test_w_limit_docratio_ge_8192_lt_16384_over_div(self):
        self.assertTrue(self._callFUT(1, 8192, 65536))
        self.assertTrue(self._callFUT(2, 8192, 65536))
        self.assertTrue(self._callFUT(4095, 8192, 65536))
        self.assertTrue(self._callFUT(4096, 8192, 65536))
        self.assertFalse(self._callFUT(4097, 8192, 65536))
        self.assertTrue(self._callFUT(1, 16383, 65536))
        self.assertTrue(self._callFUT(2, 16383, 65536))
        self.assertTrue(self._callFUT(4095, 16383, 65536))
        self.assertTrue(self._callFUT(4096, 16383, 65536))
        self.assertFalse(self._callFUT(4097, 16383, 65536))


class Test_nbest_ascending_wins(unittest.TestCase):

    def _callFUT(self, limit, rlen, numdocs):
        from zerodbext.catalog.indexes.field import nbest_ascending_wins
        return nbest_ascending_wins(limit, rlen, numdocs)

    def test_wo_limit(self):
        self.assertFalse(self._callFUT(None, 1, 1000))
        self.assertFalse(self._callFUT(None, 10, 1000))
        self.assertFalse(self._callFUT(None, 100, 1000))
        self.assertFalse(self._callFUT(None, 999, 1000))
        self.assertFalse(self._callFUT(0, 1, 1000))
        self.assertFalse(self._callFUT(0, 10, 1000))
        self.assertFalse(self._callFUT(0, 100, 1000))
        self.assertFalse(self._callFUT(0, 999, 1000))

    def test_w_limit_numdocs_le_768(self):
        self.assertTrue(self._callFUT(1, 100, 100))
        self.assertTrue(self._callFUT(10, 100, 100))
        self.assertTrue(self._callFUT(99, 100, 100))
        self.assertTrue(self._callFUT(1, 100, 767))
        self.assertTrue(self._callFUT(10, 100, 767))
        self.assertTrue(self._callFUT(99, 100, 767))
        self.assertTrue(self._callFUT(1, 100, 768))
        self.assertTrue(self._callFUT(10, 100, 768))
        self.assertTrue(self._callFUT(99, 100, 768))

    def test_w_limit_docratio_lt_4096_over_div(self):
        self.assertTrue(self._callFUT(1, 10, 65536))
        self.assertTrue(self._callFUT(2, 10, 65536))
        self.assertTrue(self._callFUT(127, 200, 65536))
        self.assertTrue(self._callFUT(128, 200, 65536))
        self.assertTrue(self._callFUT(3000, 4094, 65536))
        self.assertTrue(self._callFUT(3000, 4095, 65536))

    def test_w_limit_docratio_eq_1(self):
        self.assertTrue(self._callFUT(1, 65536, 65536))
        self.assertTrue(self._callFUT(2, 65536, 65536))
        self.assertTrue(self._callFUT(8191, 65536, 65536))
        self.assertTrue(self._callFUT(8192, 65536, 65536))
        self.assertFalse(self._callFUT(8193, 65536, 65536))

    def test_w_limit_docratio_ge_32768_over_div(self):
        self.assertTrue(self._callFUT(1, 37268, 65536))
        self.assertTrue(self._callFUT(2, 37268, 65536))
        self.assertTrue(self._callFUT(4095, 37268, 65536))
        self.assertTrue(self._callFUT(4096, 37268, 65536))
        self.assertFalse(self._callFUT(4097, 37268, 65536))
        self.assertTrue(self._callFUT(1, 65535, 65536))
        self.assertTrue(self._callFUT(2, 65535, 65536))
        self.assertTrue(self._callFUT(4095, 65535, 65536))
        self.assertTrue(self._callFUT(4096, 65535, 65536))
        self.assertFalse(self._callFUT(4097, 65535, 65536))

    def test_w_limit_docratio_ge_4096_lt_32768_over_div(self):
        self.assertTrue(self._callFUT(1, 4096, 65536))
        self.assertTrue(self._callFUT(2, 4096, 65536))
        self.assertTrue(self._callFUT(2047, 4096, 65536))
        self.assertTrue(self._callFUT(2048, 4096, 65536))
        self.assertFalse(self._callFUT(2049, 4096, 65536))
        self.assertTrue(self._callFUT(1, 32767, 65536))
        self.assertTrue(self._callFUT(2, 32767, 65536))
        self.assertTrue(self._callFUT(2047, 32767, 65536))
        self.assertTrue(self._callFUT(2048, 32767, 65536))
        self.assertFalse(self._callFUT(2049, 32767, 65536))



import unittest

class TestDocumentMap(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.document import DocumentMap
        return DocumentMap

    def _makeOne(self):
        return self._getTargetClass()()

    def test_docid_for_address_nonesuch(self):
        map = self._makeOne()
        map.add('/address1', 1)
        self.assertEqual(map.docid_for_address('/notexist'), None)

    def test_docid_for_address_existing(self):
        map = self._makeOne()
        map.add('/address1', 1)
        self.assertEqual(map.docid_for_address('/address1'), 1)

    def test_address_for_docid_nonesuch(self):
        map = self._makeOne()
        map.add('/address1', 1)
        self.assertEqual(map.address_for_docid(2), None)

    def test_address_for_docid_existing(self):
        map = self._makeOne()
        map.add('/address1', 1)
        self.assertEqual(map.address_for_docid(1), '/address1')

    def test_add_simple(self):
        map = self._makeOne()
        docid = map.add('/address')
        self.assertEqual(map.address_for_docid(docid), '/address')
        self.assertEqual(map.docid_for_address('/address'), docid)

    def test_add_with_same_address_repeatedly_replaces(self):
        map = self._makeOne()
        map.add('yup', 1)
        map.add('yup', 2)
        map.add('yup', 3)
        self.assertEqual(len(map.docid_to_address), 1)
        self.assertEqual(len(map.address_to_docid), 1)
        self.assertEqual(map.docid_to_address.get(1), None)
        self.assertEqual(map.docid_to_address.get(2), None)
        self.assertEqual(map.docid_to_address.get(3), 'yup')
        self.assertEqual(map.address_to_docid['yup'], 3)

    def test_add_existing_docid_new_address_replaces_old(self):
        map = self._makeOne()
        old_docid = map.add('/address1')
        new_docid = map.add('/address2', old_docid)
        self.assertEqual(map.address_for_docid(old_docid), '/address2')
        self.assertEqual(map.docid_for_address('/address1'), None)
        self.assertEqual(map.docid_for_address('/address2'), old_docid)
        self.assertEqual(new_docid, old_docid)

    def test_add_existing_docid_removes_metadata(self):
        map = self._makeOne()
        docid = map.add('/address1')
        map.add_metadata(docid, {'a': 1})
        map.add('/address2', docid)
        self.assertRaises(KeyError, map.get_metadata, docid)

    def test_remove_docid_nonesuch(self):
        map = self._makeOne()
        result = map.remove_docid(1)
        self.assertEqual(result, False)
        self.assertEqual(map.address_for_docid(1), None)
        self.assertEqual(map.docid_for_address('/address1'), None)

    def test_remove_docid_existing(self):
        map = self._makeOne()
        map.add('/address1', 1)
        map.add_metadata(1, {'a': 1})
        result = map.remove_docid(1)
        self.assertEqual(result, True)
        self.assertEqual(map.address_for_docid(1), None)
        self.assertEqual(map.docid_for_address('/address1'), None)
        self.assertRaises(KeyError, map.get_metadata, 1)

    def test_remove_docid_map_out_of_sync(self):
        map = self._makeOne()
        map.add('/address1', 1)
        map.add_metadata(1, {'a': 1})
        map.address_to_docid['/address1'] = 2
        result = map.remove_docid(1)
        self.assertEqual(result, True)
        self.assertEqual(map.address_for_docid(1), None)
        self.assertEqual(map.docid_for_address('/address1'), None)
        self.assertRaises(KeyError, map.get_metadata, 1)
        self.assertEqual(list(map.docid_to_address), [])
        self.assertEqual(list(map.address_to_docid), [])

    def test_remove_address_nonesuch(self):
        map = self._makeOne()
        result = map.remove_address('/address1')
        self.assertEqual(result, False)
        self.assertEqual(map.address_for_docid(1), None)
        self.assertEqual(map.docid_for_address('/address1'), None)

    def test_remove_address_exists(self):
        map = self._makeOne()
        map.add('/address1', 1)
        map.docid_to_metadata[1] = None
        result = map.remove_address('/address1')
        self.assertEqual(result, True)
        self.assertEqual(map.address_for_docid(1), None)
        self.assertEqual(map.docid_for_address('/address1'), None)
        self.assertFalse(1 in map.docid_to_metadata)

    def test_remove_address_map_out_of_sync(self):
        map = self._makeOne()
        map.add('/address1', 1)
        map.add_metadata(1, {'a': 1})
        map.docid_to_address[1] = '/address2'
        map.address_to_docid['/address2'] = 3
        result = map.remove_address('/address1')
        self.assertEqual(result, True)
        self.assertEqual(map.address_for_docid(1), None)
        self.assertEqual(map.docid_for_address('/address1'), None)
        self.assertRaises(KeyError, map.get_metadata, 1)
        self.assertEqual(list(map.docid_to_address), [])
        self.assertEqual(list(map.address_to_docid), [])


    def test_add_metadata_docid_not_in_addressmap(self):
        map = self._makeOne()
        self.assertRaises(KeyError, map.add_metadata, 1, {'a':'1'})

    def test_add_metadata_docid_in_addressmap(self):
        map = self._makeOne()
        docid = map.add('/address1')
        map.add_metadata(docid, {'a':1, 'b':2})
        self.assertEqual(map.get_metadata(docid)['a'], 1)
        self.assertEqual(map.get_metadata(docid)['b'], 2)

    def test_add_metadata_no_values(self):
        map = self._makeOne()
        docid = map.add('/address1')
        map.add_metadata(docid, {})
        self.assertFalse(docid in map.docid_to_metadata)

    def test_remove_metadata_nokeys_nodocid(self):
        map = self._makeOne()
        self.assertRaises(KeyError, map.remove_metadata, 1)

    def test_remove_metadata_nokeys_docid(self):
        map = self._makeOne()
        docid = map.add('/address1')
        map.docid_to_metadata[docid] = None
        map.remove_metadata(docid)
        self.assertFalse(docid in map.docid_to_metadata)

    def test_remove_metadata_keys_nodocid(self):
        map = self._makeOne()
        self.assertRaises(KeyError, map.remove_metadata, 1, 'a')

    def test_remove_metadata_keys_docid(self):
        map = self._makeOne()
        docid = map.add('/address1')
        map.docid_to_metadata[docid] = {'a':1, 'b':2, 'c':3}
        map.remove_metadata(docid, 'a', 'b')
        self.assertEqual(map.get_metadata(docid), {'c':3})
        map.remove_metadata(docid, 'c')
        self.assertFalse(docid in map.docid_to_metadata)

    def test_get_metadata_no_mainbtree(self):
        map = self._makeOne()
        map.docid_to_metadata = None
        self.assertRaises(KeyError, map.get_metadata, 1)

    def test_get_metadata_not_in_map(self):
        map = self._makeOne()
        self.assertRaises(KeyError, map.get_metadata, 1)

    def test_get_metadata_nokeys(self):
        map = self._makeOne()
        map.add('/address', 1)
        map.add_metadata(1, {'a':1, 'b':2})
        self.assertEqual(dict(map.get_metadata(1)), {'a':1, 'b':2})

    def test_new_docid(self):
        map = self._makeOne()
        times = [0]
        def randrange(frm, to):
            val = times[0]
            times[0] = times[0] + 1
            return val
        map._randrange = randrange
        map.add('/whatever', 0)
        self.assertEqual(map.new_docid(), 1)

    def test__check_metadata_creates_tree(self):
        map = self._makeOne()
        map.docid_to_metadata = None
        map._check_metadata()
        self.assertFalse(map.docid_to_metadata is None)
        self.assertEqual(dict(map.docid_to_metadata), {})



import six
from zope.interface import implementer

from zope.index.keyword import KeywordIndex

from zerodbext.catalog.interfaces import ICatalogIndex
from zerodbext.catalog.indexes.common import CatalogIndex


@implementer(ICatalogIndex)
class CatalogKeywordIndex(CatalogIndex, KeywordIndex):
    """
    Keyword index.

    Query types supported:

    - Eq

    - NotEq

    - In

    - NotIn

    - Any

    - NotAny

    - All

    - NotAll

    """

    def __init__(self, discriminator):
        if not callable(discriminator):
            if not isinstance(discriminator, six.string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        self._not_indexed = self.family.IF.Set()
        self.clear()

    def reindex_doc(self, docid, value):
        # the base index' index_doc method special-cases a reindex
        return self.index_doc(docid, value)

    def _indexed(self):
        return self._rev_index.keys()

    def applyAny(self, values):
        return self.apply({'query': values, 'operator': 'or'})

    applyIn = applyAny

    def applyAll(self, values):
        return self.apply({'query': values, 'operator': 'and'})

    def applyEq(self, value):
        return self.apply([value])



# 4933 sec to catalog 511421 documents with 3 field indexes, a facet index
# and a text index.

import os
import datetime

BENCHMARK_DATA_DIR='benchmark_data'
MAILLIST_INDEX='http://mail.python.org/pipermail/python-list/'

from zerodbext.catalog.catalog import FileStorageCatalogFactory
from zerodbext.catalog.catalog import ConnectionManager
from zerodbext.catalog.indexes.field import CatalogFieldIndex
from zerodbext.catalog.indexes.facet import CatalogFacetIndex
from zerodbext.catalog.indexes.text import CatalogTextIndex
        
from email.Parser import Parser
from rfc822 import parsedate_tz
import gzip,time
from urllib2 import urlopen
from HTMLParser import HTMLParser
from urlparse import urljoin

class Profiler(object):
    """This is a 'profiler' of sorts intended to let us find out how
       long particular actions, of programmer interest, take to run,
       total.  Actions can be arbitrarily nested.  This doesn't do
       anything like the actual python profiler, which tells how much
       time you end up spending in particular function calls
       aggregated across all calls to particular functions.  Both
       kinds of data are useful and recommended for getting a handle
       on where performance bottlenecks might be.
    """
    def __init__(self):
        self.action_root = TimedAction('Total')
        self.action_stack = [ self.action_root, ]
        
    def start(self, name):
        action = TimedAction(name)
        self.action_stack[-1].children.append(action)
        self.action_stack.append(action)
        print name
        
    def stop(self, name=None):
        if name is None:
            self.action_root.stop()
            return
        
        action = self.action_stack.pop()
        if action.name != name:
            raise Exception( "Profiler action stopped out of sequence.  "
                             "Expecting: %s" % action.name )
        action.stop()
        
    def print_stack(self):
        self.action_root.print_action()
        
class TimedAction(object):
    def __init__(self, name):
        self.name = name
        self.start_time = time.time()
        self.end_time = None
        self.children = []
        
    def stop(self):
        self.end_time = time.time()
        
    def print_action(self,level=0):
        indent = "  ".join( [ "" for i in xrange(level+1) ] ) # Hacky, sorry
        if self.end_time:
            print "%s%s: %0.3f" % ( indent, self.name,
                                    self.end_time - self.start_time )
        else:
            print "%s%s:" % ( indent, self.name )

        for child in self.children:
            child.print_action( level + 1 )
            
# Start profiling
profiler = Profiler()

def prep_catalog():
    """Download python mailing list, create new catalog and catalog 
       messages, if not done already.
    """
    if not os.path.exists(BENCHMARK_DATA_DIR):
        os.makedirs(BENCHMARK_DATA_DIR)
        
    # Check to see if mailing list data already present
    if len(get_mailbox_filenames()) == 0:
        MailListSucker(MAILLIST_INDEX,BENCHMARK_DATA_DIR).suck()
        
    # Create ZODB and index maillist messages, if not yet done
    zodb_file = os.path.join(BENCHMARK_DATA_DIR, 'test.zodb')
    if not os.path.exists(zodb_file):
        # Create a catalog
        manager = ConnectionManager()
        factory = FileStorageCatalogFactory(
            os.path.join(BENCHMARK_DATA_DIR,
                         'test.zodb'), 'benchmark' )
        c = factory(manager)
        
        # Create some indices
        c['subject'] = CatalogFieldIndex(get_subject)
        c['date'] = CatalogFieldIndex(get_date)
        c['sender_email'] = CatalogFieldIndex(get_sender_email)
        c['topics'] = CatalogFacetIndex(get_topics, topic_taxonomy)
        c['text'] = CatalogTextIndex(get_text)
        manager.commit()
                
        # Loop over messages to get base line
        profiler.start( "Loop over messages without indexing" )
        for _ in MessageIterator():
            pass
        profiler.stop( "Loop over messages without indexing" )
        
        profiler.start( "Index messages" )
        id = 1
        for msg in MessageIterator():
            c.index_doc(id,msg)
            id += 1
            if id / 100 == 0:
                manager.commit()
        manager.commit()
        manager.close()
        
        profiler.stop( "Index messages" )
        print "Indexed %d messages" % id
    
def get_mailbox_filenames():
    return [ dir for dir in
             os.listdir(BENCHMARK_DATA_DIR) if dir[-7:] == '.txt.gz' ]

# Adapter methods for indexing messages
def get_subject(msg,default):
    subject = msg.get('Subject',default)
    return subject

def get_date(msg,default):
    date = msg.get('Date',default)
    return date

def get_sender_email(msg,default):
    sender_email = msg.get('From', default)
    return sender_email

topic_taxonomy = set(['year'])
for year in (2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009):
    topic_taxonomy.add('year:%s' % year)
    for num in range(1, 13):
        monthname = datetime.date(year, num, 1).strftime('%B')
        topic_taxonomy.add('year:%s:%s' % (year, monthname))

def get_topics(msg, default):
    date = msg.get('Date', default)
    if date is default:
        return default
    try:
        tt = parsedate_tz(date)
    except:
        return default
    if not tt:
        return default
    year, mon, _, _, _, _, _, _, _, _ = tt
    try:
        year = int(year)
    except:
        return default
    if year < 1900:
        year = year + 1900
    monthname = datetime.date(year, mon, 1).strftime('%B')
    return ['year:%s:%s' % (year, monthname)]

def get_text(msg, default):
    return msg.as_string()

class MailListSucker(HTMLParser):
    BUFFER_SIZE = 64 * 1024
    
    def __init__(self,url,out_dir):
        HTMLParser.__init__(self)
        self.url = url
        self.out_dir = out_dir
        
    def suck(self):
        self.feed(urlopen(self.url).read())
        self.close()
        
    def blow(self):
        raise NotImplemented
    
    def handle_starttag(self,name,attrs):
        if name == 'a':
            for name, href in attrs:
                if name == 'href' and href and href[-7:] == '.txt.gz':
                    # Download file
                    href = urljoin( self.url, href )
                    print "Downloading %s..." % href
                    fname = href[href.rindex('/')+1:]
                    down = urlopen(href)
                    out = open( os.path.join( BENCHMARK_DATA_DIR,
                                              fname ), "wb" )
                    buf = down.read(self.BUFFER_SIZE)
                    while len(buf):
                        out.write(buf)
                        buf = down.read(self.BUFFER_SIZE)
                    out.close()
                    down.close()

class MessageIterator(object):
    """Iterates over a messages in a series of gzipped mailboxes in the 
       benchmark data directory.  Conveniently aggregates all messages
       in all mailboxes into a single iterable.
    """
    email_parser = Parser()
    def __init__(self):
        self.file_list = get_mailbox_filenames()
        self._next_file()
        
    def _next_file(self):
        if self.file_list:
            fname = self.file_list.pop(0)
            
            # Read whole thing into memory and manipulate it.
            # Not the most efficient but good enough for testing
            print "load %s" % fname
            self.messages = gzip.open(
                os.path.join(BENCHMARK_DATA_DIR,fname)).read().split('\nFrom ')
        
        else:    
            raise StopIteration
    
    def next(self):
        if not self.messages:
            self._next_file()
        message = self.messages.pop(0)
        return self.email_parser.parsestr( message[message.index('\n')+1:] )
        
    def __iter__(self):
        return self

def run():
    # Download mailbox archive of python mailing list and build
    # catalog if needed
    prep_catalog()

    # Open a catalog
    manager = ConnectionManager()
    factory = FileStorageCatalogFactory(
        os.path.join(BENCHMARK_DATA_DIR,'test.zodb'), 'benchmark' )
    c = factory(manager)

    # Do some searches

    profiler.start( "unsorted retrieval" )
    n, results = c.search(date=('0', 'Z'))
    print '%d results ' % n
    # Force generator to marshall brains
    for result in results:
        pass
    profiler.stop( "unsorted retrieval" )

    profiler.start( "repeat unsorted retrieval" )
    n, results = c.search(date=('0', 'Z'))
    print '%d results ' % n
    # Force generator to marshall brains
    for result in results:
        pass
    profiler.stop( "repeat unsorted retrieval" )

    profiler.start( "sorted retrieval" )
    n, results = c.search( date=('0', 'Z'), sort_index='subject' )
    print '%d results ' % n
    for result in results:
        pass
    profiler.stop( "sorted retrieval" )

    profiler.start( "reverse sorted retrieval" )
    n, results = c.search( date=('0', 'Z'), sort_index='subject', reverse=True )
    print '%d results ' % n
    for result in results:
        pass
    profiler.stop( "reverse sorted retrieval" )

    profiler.start('limit to topic=year:2000')
    n, results = c.search( topics=['year:2000'] )
    print '%d results' % n
    L = []
    for result in results:
        L.append(result)
    profiler.stop( "limit to topic=year:2000" )

    profiler.start('count limited to topic=year:2000')
    print c['topics'].counts(L, ['year:2000'])
    profiler.stop('count limited to topic=year:2000')

    profiler.stop()
    profiler.print_stack()




import six
import BTrees
from persistent.mapping import PersistentMapping
import transaction

from zope.interface import implementer

from zerodbext.catalog.interfaces import ICatalog
from zerodbext.catalog.interfaces import ICatalogIndex


@implementer(ICatalog)
class Catalog(PersistentMapping):

    family = BTrees.family32

    def __init__(self, family=None):
        PersistentMapping.__init__(self)
        if family is not None:
            self.family = family

    def clear(self):
        """ Clear all indexes in this catalog. """
        for index in self.values():
            index.clear()

    def index_doc(self, docid, obj):
        """Register the document represented by ``obj`` in indexes of
        this catalog using docid ``docid``."""
        assertint(docid)
        for index in self.values():
            index.index_doc(docid, obj)

    def unindex_doc(self, docid):
        """Unregister the document id from indexes of this catalog."""
        assertint(docid)
        for index in self.values():
            index.unindex_doc(docid)

    def reindex_doc(self, docid, obj):
        """ Reindex the document referenced by docid using the object
        passed in as ``obj`` (typically just does the equivalent of
        ``unindex_doc``, then ``index_doc``, but specialized indexes
        can override the method that this API calls to do less work. """
        assertint(docid)
        for index in self.values():
            index.reindex_doc(docid, obj)

    def __setitem__(self, name, index):
        """ Add an object which implements
        ``zerodbext.catalog.interfaces.ICatalogIndex`` to the catalog.
        No other type of object may be added to a catalog."""
        if not ICatalogIndex.providedBy(index):
            raise ValueError('%s does not provide ICatalogIndex')
        return PersistentMapping.__setitem__(self, name, index)

    def search(self, **query):
        """ Use the query terms to perform a query.  Return a tuple of
        (num, resultseq) based on the merging of results from
        individual indexes.

        .. note::

                  this method is deprecated as of :mod:`zerodbext.catalog`
                  version 0.8.  Use :meth:`zerodbext.catalog.Catalog.query`
                  instead.


        """
        sort_index = None
        reverse = False
        limit = None
        sort_type = None
        index_query_order = None

        if 'sort_index' in query:
            sort_index = query.pop('sort_index')
        if 'reverse' in query:
            reverse = query.pop('reverse')
        if 'limit' in query:
            limit = query.pop('limit')
        if 'sort_type' in query:
            sort_type = query.pop('sort_type')
        if 'index_query_order' in query:
            index_query_order = query.pop('index_query_order')

        if index_query_order is None:
            # unordered query (use apply)
            results = []
            for index_name, index_query in query.items():
                index = self.get(index_name)
                if index is None:
                    raise ValueError('No such index %s' % index_name)
                r = index.apply(index_query)
                if not r:
                    # empty results, bail early; intersect will be null
                    return EMPTY_RESULT

                results.append((len(r), r))

            if not results:
                return EMPTY_RESULT

            results.sort(key=len) # order from smallest to largest
            _, result = results.pop(0)
            for _, r in results:
                _, result = self.family.IF.weightedIntersection(result, r)

            if not result:
                EMPTY_RESULT

        else:
            # ordered query (use apply_intersect)
            result = None
            _marker = object()
            for index_name in index_query_order:
                index_query = query.get(index_name, _marker)
                if index_query is _marker:
                    continue
                index = self.get(index_name)
                if index is None:
                    raise ValueError('No such index %s' % index_name)
                result = index.apply_intersect(index_query, result)
                if not result:
                    # empty results
                    return EMPTY_RESULT

        return self.sort_result(result, sort_index, limit, sort_type, reverse)

    def sort_result(self, result, sort_index=None, limit=None, sort_type=None,
                    reverse=False):

        numdocs = total = len(result)

        if sort_index:
            index = self[sort_index]
            result = index.sort(result, reverse=reverse, limit=limit,
                                sort_type=sort_type)
            if limit:
                numdocs = min(numdocs, limit)
        return ResultSetSize(numdocs, total), result

    def query(self, queryobject, sort_index=None, limit=None, sort_type=None,
              reverse=False, names=None):
        """ Use the arguments to perform a query.  Return a tuple of
        (num, resultseq)."""
        try:
            from zerodbext.catalog.query import parse_query
            if isinstance(queryobject, six.string_types):
                queryobject = parse_query(queryobject)
        except ImportError: #pragma NO COVERAGE
            pass
        results = queryobject._apply(self, names)
        return self.sort_result(results, sort_index, limit, sort_type, reverse)

    def apply(self, query):
        return self.search(**query)

def assertint(docid):
    if not isinstance(docid, int):
        raise ValueError('%r is not an integer value; document ids must be '
                         'integers' % docid)

class CatalogFactory(object):
    def __call__(self, connection_handler=None):
        conn = self.db.open()
        if connection_handler:
            connection_handler(conn)
        root = conn.root()
        if root.get(self.appname) is None:
            root[self.appname] = Catalog()
        return root[self.appname]

class FileStorageCatalogFactory(CatalogFactory):
    def __init__(self, filename, appname, **kw):
        """ ``filename`` is a filename to the FileStorage storage,
        ``appname`` is a key name in the root of the FileStorage in
        which to store the catalog, and ``**kw`` is passed as extra
        keyword arguments to :class:`ZODB.DB.DB` when creating a
        database.  Note that when we create a :class:`ZODB.DB.DB`
        instance, if a ``cache_size`` is not passed in ``*kw``, we
        override the default ``cache_size`` value with ``50000`` in
        order to provide a more realistic cache size for modern apps"""
        cache_size = kw.get('cache_size')
        if cache_size is None:
            kw['cache_size'] = 50000

        from ZODB.FileStorage.FileStorage import FileStorage
        from ZODB.DB import DB
        f = FileStorage(filename)
        self.db = DB(f, **kw)
        self.appname = appname

    def __del__(self):
        self.db.close()

class ConnectionManager(object):
    def __call__(self, conn):
        self.conn = conn

    def close(self):
        self.conn.close()

    def __del__(self):
        self.close()

    def commit(self, transaction=transaction):
        transaction.commit()

class ResultSetSizeClass(int):

    def __repr__(self):
        return 'ResultSetSize(%d, %d)' % (self, self.total)

def ResultSetSize(i, total):
    size = ResultSetSizeClass(i)
    size.total = total
    return size

EMPTY_RESULT = ResultSetSize(0, 0), ()
